"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import unittest

from hyram.qra import event_tree


class TestEventTreeCalcs(unittest.TestCase):
    """
    Tests of calculations for event trees
    """
    def test_negative_probability_value_error(self):
        self.assertRaises(ValueError, event_tree.check_probability_value, -1)

    def test_probability_value_too_large_error(self):
        self.assertRaises(ValueError, event_tree.check_probability_value, 1.1)

    def test_probability_not_occur(self):
        probability = 0.7
        probability_not_occur = 0.3
        calculated_not_probability = event_tree.calc_probability_not_occur(probability)
        self.assertAlmostEqual(calculated_not_probability, probability_not_occur)

    def test_calc_end_state_frequencies(self):
        initiating_event_frequency = 1e-5
        end_state_probabilities = [0.1, 0.2, 0.3, 0.4]
        end_state_frequencies = event_tree.calc_end_state_frequencies(initiating_event_frequency, end_state_probabilities)
        hand_calc_end_state_freqs = [1e-6, 2e-6, 3e-6, 4e-6]
        for calc_value, hand_calc_value in zip(end_state_frequencies, hand_calc_end_state_freqs):
            self.assertAlmostEqual(calc_value, hand_calc_value)


class TestBuildEventTree(unittest.TestCase):
    """
    Tests of building an event tree calculation
    """
    def test_calc_end_state_probabilities(self):
        event_probabilities = [0.1, 0.2, 0.3]
        end_state_probabilities = event_tree.calc_end_state_probabilities(event_probabilities)
        hand_calc_end_state_probabilities = [0.1, 0.18, 0.216, 0.504]
        for calc_value, hand_calc in zip(end_state_probabilities, hand_calc_end_state_probabilities):
            self.assertAlmostEqual(calc_value, hand_calc)

    def test_end_state_probabilities_sum_to_1(self):
        event_probabilities = [0.1, 0.2, 0.3]
        end_state_probabilities = event_tree.calc_end_state_probabilities(event_probabilities)
        sum_of_end_state_probabilities = sum(end_state_probabilities)
        self.assertAlmostEqual(sum_of_end_state_probabilities, 1)

    def test_reject_bad_probabilities(self):
        bad_event_probabilities = [0.1, 1.2, 0.3]
        self.assertRaises(ValueError, event_tree.calc_end_state_probabilities, bad_event_probabilities)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""
import unittest

import scipy.constants as spc

from hyram.qra import analysis as qra_analysis
from hyram.qra import component_failure
from hyram.qra import component_set
from hyram.qra.data import component_data


class HydrogenTestCase(unittest.TestCase):
    """
    Test calculation of QRA analysis
    """
    def setUp(self):
        rel_species = "h2"
        rel_phase = None
        failure_set = component_failure.ComponentFailureSet(f_failure_override=None,
                                                            num_vehicles=20, daily_fuelings=2, vehicle_days=250,
                                                            noz_po_dist='beta', noz_po_a=0.5, noz_po_b=610415.5,
                                                            noz_ftc_dist='expv', noz_ftc_a=0.002, noz_ftc_b=None,
                                                            mvalve_ftc_dist='expv', mvalve_ftc_a=0.001, mvalve_ftc_b=None,
                                                            svalve_ftc_dist='expv', svalve_ftc_a=0.002, svalve_ftc_b=None,
                                                            svalve_ccf_dist='expv', svalve_ccf_a=0.000128, svalve_ccf_b=None,
                                                            overp_dist='beta', overp_a=3.5, overp_b=310289.5,
                                                            pvalve_fto_dist='logn', pvalve_fto_a=-11.74, pvalve_fto_b=0.67,
                                                            driveoff_dist='beta', driveoff_a=31.5, driveoff_b=610384.5,
                                                            coupling_ftc_dist='beta', coupling_ftc_a=0.5, coupling_ftc_b=5031)
        self.component_sets = component_set.init_component_sets(species=rel_species, phase=rel_phase,
                                                                n_valves=5, n_instruments=3, n_joints=35, n_hoses=1, n_pipes=20)
        occupant_input_list = [{
            'count': 9,
            'descrip': 'workers',
            'xdistr': 'uniform', 'xa': 1, 'xb': 20,
            'ydistr': 'dete', 'ya': 1, 'yb': None,
            'zdistr': 'unif', 'za': 1, 'zb': 12,
            'hours': 2000,
        }]

        self.params = {
            "pipe_outer_diam": 0.375 * spc.inch,  # m
            "pipe_thickness": 0.065 * spc.inch,  # m
            "amb_temp": 15 + spc.zero_Celsius,  # K
            "amb_pres": 101325,  # Pa
            "rel_temp": 15 + spc.zero_Celsius,  # K
            "rel_pres": 35  * spc.mega,  # Pa
            "rel_phase": None,
            "facil_length": 20,  # m
            "facil_width": 12,  # m
            "immed_ign_probs": [0.008, 0.053, 0.23],
            "delayed_ign_probs": [0.004, 0.027, 0.12],
            "ign_thresholds": [0.125, 6.25],
            "occupant_input_list": occupant_input_list,
            "component_sets": self.component_sets,
            "component_failure_set": failure_set,
            "leak_sizes": None,
            "rel_species": 'h2',
            "discharge_coeff": 1,
            "detection_credit": 0.9,
            "overp_method": 'bst',
            "tnt_factor": None,
            "bst_flame_speed": 0.35,
            "probit_thermal_id": 'eise',
            "exposure_time": 60,  # s
            "probit_overp_id": 'head',
            "nozzle_model": 'yuce',
            "rel_angle": 0,
            "rel_humid": 0.89,
            "rand_seed": 3632850,
            "excl_radius": 0.01,
            "f_release_overrides": None,
            "verbose": False,
            "output_dir": None,  # Use default "temp"
            "create_plots": False,
        }

    def test_zero_risk_for_zero_components_zero_refuelings(self):
        zero_component_sets = []
        for component_set in self.component_sets:
            component_set.num_components = 0
            zero_component_sets.append(component_set)
        zero_refuelings_component_failure_set = component_failure.ComponentFailureSet(num_vehicles=0)
        self.params['component_sets'] = zero_component_sets
        self.params['component_failure_set'] = zero_refuelings_component_failure_set

        results = qra_analysis.conduct_analysis(**self.params)
        risk_value = results['total_pll']
        self.assertAlmostEqual(risk_value, 0)

    def test_zero_risk_for_zero_occupants(self):
        zero_occupant_input_list = [{
            'count': 0,
            'descrip': 'workers',
            'xdistr': 'uniform', 'xa': 1, 'xb': 20,
            'ydistr': 'dete', 'ya': 1, 'yb': None,
            'zdistr': 'unif', 'za': 1, 'zb': 12,
            'hours': 2000,
        }]
        self.params['occupant_input_list'] = zero_occupant_input_list
        results = qra_analysis.conduct_analysis(**self.params)
        risk_value = results['total_pll']
        self.assertAlmostEqual(risk_value, 0)

    def test_zero_risk_for_100pct_detection(self):
        detection_credit_100pct = 1
        self.params['detection_credit'] = detection_credit_100pct
        results = qra_analysis.conduct_analysis(**self.params)

        risk_value = results['total_pll']
        self.assertAlmostEqual(risk_value, 0)

    def test_reject_bauwens_with_impulse_probits(self):
        for probit_overp_id in ['head', 'coll']:
            self.params['overp_method'] = 'bauwens'
            self.params['probit_overp_id'] = probit_overp_id
            with self.assertRaises(ValueError):
                results = qra_analysis.conduct_analysis(**self.params)

    def test_flexible_leak_size_specification(self):
        self.params['leak_sizes'] = [0.1, 10]
        results1 = qra_analysis.conduct_analysis(**self.params)
        risk_value1 = results1['total_pll']

        self.params['leak_sizes'] = [100, 10, 0.1]
        results2 = qra_analysis.conduct_analysis(**self.params)
        risk_value2 = results2['total_pll']

        self.assertEqual(risk_value1, risk_value2)
        self.assertEqual(len(results1['leak_results']), len(results2['leak_results']))

    def test_overriding_leak_freq(self):
        self.params['f_release_overrides'] = [0, 0, 0, 0, 0]
        self.params['component_failure_set'] = component_failure.ComponentFailureSet(f_failure_override=0.0)
        results = qra_analysis.conduct_analysis(**self.params)
        risk_value = results['total_pll']
        self.assertEqual(risk_value, 0)

    def test_sat_vapor(self):
        self.params['rel_phase'] = 'gas'
        self.params['rel_temp'] = None
        self.params['rel_pres'] = 1e6
        self.params['component_sets'] = component_set.init_component_sets(species='h2', phase='gas',
                                                                          n_vessels=1, n_flanges=1, n_hoses=1,
                                                                          n_joints=35, n_pipes=20, n_valves=5, n_instruments=0,
                                                                          fuel_component_data=component_data.h2_liquid_params)
        results = qra_analysis.conduct_analysis(**self.params)
        risk_value = results['total_pll']
        self.assertAlmostEqual(risk_value, 0)

    def test_sat_liquid(self):
        self.params['rel_phase'] = 'liquid'
        self.params['rel_temp'] = None
        self.params['rel_pres'] = 1e6
        self.params['component_sets'] = component_set.init_component_sets(species='h2', phase='gas',
                                                                          n_vessels=1, n_flanges=1, n_hoses=1,
                                                                          n_joints=35, n_pipes=20, n_valves=5, n_instruments=0,
                                                                          fuel_component_data=component_data.h2_liquid_params)
        results = qra_analysis.conduct_analysis(**self.params)
        risk_value = results['total_pll']
        self.assertAlmostEqual(risk_value, 0)


if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import numpy as np
from scipy.stats import norm

from ..utilities import misc_utils


def calculate_fatality_probability(probit_value, normal_distribution_mean=5):
    """
    Calculate probability of fatality from probit value

    Parameters
    ----------
    probit_value : float
        Value from probit model
    normal_distribution_mean : int
        Default value of 5 avoids negative values, consistent with published models

    Returns
    -------
    prob : float
        Probability of fatality

    """
    return norm.cdf(probit_value, loc=normal_distribution_mean)


"""
THERMAL Fatality Probit Models
"""

def compute_thermal_fatality_prob(model_ref, heat_flux, exposure_time, mean=5):
    """
    Calculate probability of fatality from thermal source using user-specified model

    Parameters
    ----------
    model_ref : str
        reference to internal thermal probit model function to apply (see below)
    heat_flux : float
        heat flux intensity (W/m^2)
    exposure_time : float
        duration of exposure (s)
    mean : int
        Default value of 5 avoids negative values, consistent with published models

    Returns
    -------
    prob : float
        Probability of fatality
    """
    cleaned_id = parse_thermal_model(model_ref)
    probit_model = PROBIT_THERMAL_CHOICES[cleaned_id]

    thermal_dose = calculate_thermal_dose(heat_flux, exposure_time)
    if thermal_dose == 0:
        probit_value = -np.inf
    else:
        probit_value = probit_model(thermal_dose)

    return calculate_fatality_probability(probit_value, mean)


def calculate_thermal_dose(heat_flux, exposure_time):
    """
    Compute thermal dose for given time

    Parameters
    -------------
    heat_flux : float
        heat flux intensity (W/m^2)
    exposure_time : float
        duration of exposure (s)

    Returns
    ---------
    thermal_dose : float
        Thermal dose in (W/m^2)^4/3 s
    """
    return exposure_time * heat_flux ** (4/3)


def thermal_eisenberg(thermal_dose):
    """
    Eisenberg - thermal exposure

    Parameters
    -------------
    thermal_dose : float
        Thermal dose in (W/m^2)^4/3 s

    Returns
    ---------
    probit_value : float
    """
    return -38.48 + 2.56 * np.log(thermal_dose)


def thermal_tsao_perry(thermal_dose):
    """
    Tsao & Perry - thermal exposure

    Parameters
    -------------
    thermal_dose : float
        Thermal dose in (W/m^2)^4/3 s

    Returns
    ---------
    probit_value : float
    """
    return -36.38 + 2.56 * np.log(thermal_dose)


def thermal_tno(thermal_dose):
    """
    TNO - thermal exposure

    Parameters
    -------------
    thermal_dose : float
        Thermal dose in (W/m^2)^4/3 s

    Returns
    ---------
    probit_value : float
    """
    return -37.23 + 2.56 * np.log(thermal_dose)


def thermal_lees(thermal_dose):
    """
    Lees - thermal exposure

    Parameters
    -------------
    thermal_dose : float
        Thermal dose in (W/m^2)^4/3 s

    Returns
    ---------
    probit_value : float
    """
    return -29.02 + 1.99 * np.log(0.5 * thermal_dose)



"""
OVERPRESSURE Fatality Probit Models

TNO Lung hemorrhage requires the person's mass, so we don't calculate it

TNO whole body impact model produces lower probabilities than head
impact model, which means head impact fatalities will dominate whole
body impact fatalities. So we don't include TNO whole body fatalities.
"""

def compute_overpressure_fatality_prob(model_ref, overp, impulse=None, mean=5):
    """
    Calculate probability of fatality from overpressure using user-specified model

    Parameters
    ----------
    model_ref : str
        reference to internal probit model function to use (from available functions)
    overp : float
        Peak overpressure (Pa)
    impulse : float
        Impulse of shock wave (Pa*s)
        Not used in Eisenberg - Lung hemorrhage or HSE - Lung Hemorrhage models
    mean : int
        Default value of 5 avoids negative values, consistent with published models

    Returns
    -------
    prob : float
        Probability of fatality
    """
    cleaned_id = parse_overp_model(model_ref)
    probit_model = PROBIT_OVERP_CHOICES[cleaned_id]

    if overp == 0 or impulse == 0:
        probit_value = -np.inf
    else:
        if cleaned_id in ['leis', 'lhse']:
            probit_value = probit_model(overp=overp)
        else:
            probit_value = probit_model(overp=overp, impulse=impulse)

    return calculate_fatality_probability(probit_value, mean)


def overp_eisenberg(overp):
    """
    Eisenberg - Lung hemorrhage

    Parameters
    ----------
    overp : float
        Peak overpressure (Pa)

    Returns
    -------
    probit_value : float
    """
    return -77.1 + 6.91 * np.log(overp)


def overp_hse(overp):
    """
    HSE - Lung hemorrhage

    Parameters
    ----------
    overp : float
        Peak overpressure (Pa)

    Returns
    -------
    probit_value : float
    """
    return 5.13 + 1.37 * np.log(overp * 1e-5)


def overp_tno_head(overp, impulse):
    """
    TNO - Head impact

    Parameters
    ----------
    overp : float
        Peak overpressure (Pa)
    impulse : float
        Impulse of shock wave (Pa*s)

    Returns
    -------
    probit_value : float
    """
    return 5 - 8.49 * np.log((2430 / overp) + 4.e8 / (overp * impulse))


def overp_tno_struct_collapse(overp, impulse):
    """
    TNO - Structural collapse

    Parameters
    ----------
    overp : float
        Peak overpressure (Pa)
    impulse: impulse of shock wave (Pa*s)

    Returns
    -------
    probit_value : float
    """
    return 5 - 0.22 * np.log((40000 / overp) ** 7.4 + (460 / impulse) ** 11.3)



# Reference to fatality models
#   key: internal reference string
#   val: function reference
PROBIT_THERMAL_CHOICES = {
    'eise': thermal_eisenberg,
    'tsao': thermal_tsao_perry,
    'tno': thermal_tno,
    'lees': thermal_lees,
}

# Reference to overpressure models
#   key: internal reference string
#   val: function reference
# Note: keys are unique with above dict to ensure no accidental overlap
PROBIT_OVERP_CHOICES = {
    'leis': overp_eisenberg,
    'lhse': overp_hse,
    'head': overp_tno_head,
    'coll': overp_tno_struct_collapse,
}


def parse_thermal_model(name):
    """ Determine model ID from string name """
    cleaned = misc_utils.clean_name(name)  # alphanumeric lower-case

    if cleaned in ['eise', 'eisenberg', 'eis', 'eisen']:
        model_id = 'eise'
    elif cleaned in ['tsao', 'tsa']:
        model_id = 'tsao'
    elif cleaned in ['tno', 'tn']:
        model_id = 'tno'
    elif cleaned in ['lees', 'lee', 'le']:
        model_id = 'lees'
    else:
        raise ValueError("Thermal model name {} not recognized".format(cleaned))

    return model_id


def parse_overp_model(name):
    """ Determine model ID from string name """
    cleaned = misc_utils.clean_name(name)  # alphanumeric lower-case

    if cleaned in ['leis', 'lung_eisenberg', 'lungeisenberg', 'lunge', 'lung_eis', 'elh']:
        model_id = 'leis'
    elif cleaned in ['lhse', 'lung_hse', 'lunghse', 'lungh', 'lhs']:
        model_id = 'lhse'
    elif cleaned in ['head_impact', 'head', 'headimpact', 'hea']:
        model_id = 'head'
    elif cleaned in ['col', 'collapse', 'coll']:
        model_id = 'coll'
    else:
        raise ValueError("Probit overpressure model name {} not recognized".format(cleaned))

    return model_id



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import numpy as np

from . import effects, ignition_probs, pipe_size, risk, event_tree, consequence, leak_size_results
from . import component_set
from . import positions as qra_positions
from ..phys import api as phys_api
from ..phys import _comps, _jet, _therm
from ..utilities import misc_utils


def conduct_analysis(pipe_outer_diam, pipe_thickness,
                     amb_temp, amb_pres,
                     rel_temp, rel_pres, rel_phase,
                     facil_length, facil_width,
                     immed_ign_probs,
                     delayed_ign_probs,
                     ign_thresholds,
                     occupant_input_list,
                     component_sets,
                     component_failure_set,
                     rel_species,
                     leak_sizes=None,
                     discharge_coeff=1, detection_credit=0.9,
                     overp_method='bst',
                     tnt_factor=None,
                     bst_flame_speed=None,
                     probit_thermal_id='eise', exposure_time=30, probit_overp_id='head',
                     nozzle_model='yuce',
                     rel_angle=0, rel_humid=0.89,
                     rand_seed=None, excl_radius=0.01,
                     mass_flow=None, mass_flow_leak_size=None,
                     f_release_overrides=None,
                     verbose=False,
                     output_dir=None,
                     create_plots=True):
    """
    Quantitative risk assessment including scenario calculations and harm modeling

    Default values for optional overrides (to not use override) is -1 due to type restrictions from C# calls

    Parameters
    ----------
    pipe_outer_diam : float
        [m] Outer diameter of pipe

    pipe_thickness : float
        [m] Thickness of pipe wall (single side)

    amb_temp : float
        [K] Ambient temperature

    amb_pres : float
        [Pa] Ambient pressure (absolute)

    rel_temp : float
        [K] Fluid temperature

    rel_pres : float
        [Pa] Fluid pressure (absolute)

    rel_phase : {'gas', 'liquid', None}
        Fluid phase; gas implies saturated vapor, liquid implies saturated liquid.
        None corresponds to default 'gas' in GUI.

    facil_length : float
        [m] Length of facility

    facil_width : float
        [m] Width of facility

    immed_ign_probs : list of floats
        List of immediate ignition probabilities based on the mass flow rate thresholds 

    delayed_ign_probs : list of floats
        List of delayed ignition probabilities based on the mass flow rate thresholds

    ign_thresholds : list of floats
        Sorted list of mass flow rates [kg/s] that separate the list of ignition probabilities

    occupant_input_list : list of dicts
        Each dict defines group of occupants/workers near radiative source.
        Format: {count, descrip, xdistr, xa, xb, ydistr, ya, yb, zdistr, za, zb, hours}
        Where distributions can be uniform, deterministic or normal, {'unif', 'dete', 'norm'}.
        Example:
            {
                'count': 9,
                'descrip': 'workers',
                'xdistr': 'uniform', 'xa': 1, 'xb': 20,
                'ydistr': 'dete', 'ya': 1, 'yb': None,
                'zdistr': 'unif', 'za': 1, 'zb': 12,
                'hours': 2000,
            }

    component_sets : [ComponentSet]
        List of components (e.g. compressors)

    component_failure_set : ComponentFailureSet
        Object representing component failure properties and parameters.

    rel_species : string or dict
        Release fluid species (e.g., 'h2', 'ch4') or dict of species and concentrations

    leak_sizes : [floats] or None
        List of percentages representing % leak

    discharge_coeff : float
        [-] Discharge coefficient to account for non-plug flow (always <=1, assumed to be 1 for plug flow)

    detection_credit : float
        Probability of detecting flame/release, as fraction

    overp_method : {'tnt', 'bst', 'bauwens'}
        ID of unconfined overpressure model to use

    tnt_factor : float
        TNT mass equivalence factor for use with TNT unconfined overpressure model
        (unused if TNT overpressure model is not used)

    bst_flame_speed : float
        One of {0.2, 0.35, 0.7, 1.0, 1.4, 2.0, 3.0, 4.0, 5.2}
        Mach flame speed for use with BST unconfined overpressure model
        (unused if BST overpressure model is not used)

    probit_thermal_id : {'eise', 'tsao', 'tno', 'lees'}
        ID of thermal probit model to use

    exposure_time : float
        [s] Duration of exposure to heat source

    probit_overp_id : {'leis', 'lhse', 'head', 'coll'}
        ID of overpressure probit model to use

    nozzle_model : {'yuce', 'ewan', 'birc', 'bir2', 'molk'}
        4-char key referencing notional nozzle model to use for high-pressure release

    rel_angle : float
        [deg] Leak release angle for use in qrad flame calculation

    rel_humid : float
        Relative humidity between 0.0 and 1.0

    rand_seed : int or None
        Random seed value for random number position generator.
        If None, will generate new random seed each call.

    mass_flow : float, optional
        [kg/s] Mass flow rate for specified leak size, if unchoked

    mass_flow_leak_size : {1, 10, 100, 1000, 10000}, optional
        Represents leak size (as hundredth of percent) of specified mass flow rate, assuming unchoked flow.

    excl_radius : float
        [m] Exclusion radius, for use in qrad flame calculation

    f_release_overrides : list of floats or None
        Manual override values for release frequency at each leak size.
        Not used if == -1 or None
        If None, vals will be set to -1 (i.e. ignored).

    verbose : bool
        If True, extra output will be printed (default False)

    output_dir : str
        Path of directory for saving local files,
        e.g., log-file and plot images

    create_plots : bool
        Whether output plots should be created

    Returns
    -------
    results : dict
        Compilation of analysis results containing:
            air : float
                Average Individual Risk is expected # of fatalities per exposed individual
            far : float
                Fatal Accident Rate is expected # of fatalities per 100 million exposed hours
            total_pll : float
                Potential Loss of Life is expected # of fatalities per system year
            qrad_plot_files : list of strings
                File locations of QRAD plots for each leak size, in order
            overp_plot_files : list of strings
                File locations of overpressure plots for each leak size, in order
            impulse_plot_files : list of strings
                File locations of impulse plots for each leak size, in order
            leak_results : list of LeakResult objects
                Each contains PLL contribution, expected probabilities for scenarios,
                and component leak probabilities
            positions : 2d array
                (x,y,z) coordinates of occupants
            position_qrads : 2d array
                flux data [W/m2] per leak per position
                e.g. for 9 positions with 5 leak sizes, 9x5 array
            position_overps : 2d array
                peak overpressure data [Pa] per leak per position
                e.g. for 9 positions with 5 leak sizes, 9x5 array
            position_impulses : 2d array
                impulse data [Pa s] per leak per position,
                e.g. for 9 positions with 5 leak sizes, 9x5 array
    """
    if probit_overp_id in ['head', 'coll'] and overp_method == 'bauwens':
        impulse_probit_error_msg = ('Overpressure method "bauwens"'
                                    + ' does not produce impulse values,'
                                    + ' and so cannot be used with'
                                    + f' overpressure probit "{probit_overp_id}"')
        raise ValueError(impulse_probit_error_msg)

    params = locals()
    sorted_params = sorted(params)

    if output_dir is None:
        output_dir = misc_utils.get_temp_folder()

    if verbose:
        print("=== BEGINNING ANALYSIS ===")
        print("Parameters:")
        for param_name in sorted_params:
            param_val = params[param_name]
            if isinstance(param_val, list) and isinstance(param_val[0], component_set.ComponentSet):
                print("Component Sets:")
                for comp_set in param_val:
                    print(f"{comp_set}")
            else:
                print("{}: {}".format(param_name, str(param_val)))
        print("\n")

    amb_fluid = phys_api.create_fluid('AIR', amb_temp, amb_pres)
    rel_fluid = phys_api.create_fluid(rel_species, rel_temp, rel_pres, phase=rel_phase)
    if rel_temp is None:
        rel_temp = rel_fluid.T
    if amb_temp is None:
        amb_temp = amb_fluid.T

    # Each occupant row in GUI is represented as group and stored as dict inside list
    # Massage into required format for phys module [count, (xdistr, xa, xb), (ydistr...]
    loc_distributions = []
    total_occupants = 0
    total_occupant_hours = 0
    for group_dict in occupant_input_list:
        num_occupants = int(group_dict['count'])
        loc_distribution = [num_occupants,
                            (group_dict['xdistr'], group_dict['xa'], group_dict['xb']),
                            (group_dict['ydistr'], group_dict['ya'], group_dict['yb']),
                            (group_dict['zdistr'], group_dict['za'], group_dict['zb'])]
        loc_distributions.append(loc_distribution)
        total_occupants += num_occupants
        total_occupant_hours += int(group_dict['hours'] * num_occupants)
    if total_occupant_hours == 0:
        occupant_avg_hours = 0
    else:
        occupant_avg_hours = total_occupant_hours / total_occupants

    if verbose:
        component_set_log_msg = ""
        for compset in component_sets:
            if compset.num_components:
                component_set_log_msg += "{}\n".format(str(compset))
        print(component_set_log_msg)
        print(f"Location distributions: {loc_distributions}")
        print(f"{total_occupants} Occupants for {occupant_avg_hours} average hours\n")

    leak_results = leak_size_results.init_leak_results(leak_sizes, f_release_overrides, component_sets,
                                                       mass_flow_leak_size=mass_flow_leak_size, mass_flow=mass_flow)
    leak_result100 = leak_results[-1]
    num_leak_sizes = len(leak_results)

    # Account for non-leak fueling failure contributors in 100% release only. Will use override value if provided.
    leak_result100.set_failure_set(component_failure_set)
    total_leak_freqs = np.array([leak_res.f_release for leak_res in leak_results])

    # Compute leak diameters and discharge rates, one per leak size
    pipe_inner_diam = pipe_size.calc_pipe_inner_diameter(pipe_outer_diam, pipe_thickness)
    pipe_flow_area = pipe_size.calc_pipe_flow_area(pipe_inner_diam)

    if verbose:
        print("RELEASE FREQUENCIES:")
        for leak_result in leak_results:
            print(f" {leak_result.leak_size:0.2f}% - {leak_result.f_release:.3g}")
        print(f"System pipe inner diameter {pipe_inner_diam:.3g} m, area {pipe_flow_area:.3g} m^2")

    orifices = []
    discharge_rates = []
    developing_flows = []
    cons_momentum, notional_noz_t = misc_utils.convert_nozzle_model_to_params(nozzle_model, rel_fluid)
    # Calculate mass flows for leak sizes if given unchoked mass flow for single leak size

    for leak_result in leak_results:
        orifice_leak_diam = pipe_size.calc_orifice_diameter(pipe_flow_area, leak_result.leak_size/100)
        orifice = _comps.Orifice(orifice_leak_diam, discharge_coeff)
        orifices.append(orifice)
        developing_flow = _jet.DevelopingFlow(rel_fluid, orifice, amb_fluid, theta0=rel_angle,
                                              nn_conserve_momentum=cons_momentum, nn_T=notional_noz_t, verbose=verbose)
        developing_flows.append(developing_flow)
        # discharge_rate = orifice.mdot(developing_flow.fluid_orifice)
        discharge_rate = orifice.mdot(orifice.flow(rel_fluid, mdot=leak_result.mass_flow_override))
        discharge_rates.append(discharge_rate)
        if verbose:
            print(f"For {leak_result.leak_size}% leak size: orifice leak diameter: {orifice_leak_diam:.3g} m, "
                     f"discharge rate: {discharge_rate:.3g} kg/s")

    # Evaluate event tree
    # Determine ignition probabilities for each leak size based on discharge rates and thresholds
    end_state_probabilities_per_leak = []
    end_state_consequence_types = {'Shutdown': None,
                                   'No Ignition': None,
                                   'Jet Fire': 'thermal',
                                   'Explosion': 'overp'}
    for rate in discharge_rates:
        (immed_ign_prob, delayed_ign_prob) = ignition_probs.get_ignition_probability(rate,
                                                                                     ign_thresholds,
                                                                                     immed_ign_probs,
                                                                                     delayed_ign_probs)
        if verbose:
            print("Flow rate {:.3g} (kg/s) ignition probabilities: immed {}, delayed {}".format(rate, immed_ign_prob, delayed_ign_prob))

        # Calculate event tree end-state probabilities
        total_ign_prob = immed_ign_prob + delayed_ign_prob
        no_ign_prob = event_tree.calc_probability_not_occur(total_ign_prob)
        conditional_immed_ign_prob = immed_ign_prob / total_ign_prob
        event_probabilities = [detection_credit, no_ign_prob, conditional_immed_ign_prob]
        end_state_probabilities = event_tree.calc_end_state_probabilities(event_probabilities)
        end_state_probabilities_per_leak.append(end_state_probabilities)

    # Generate positions
    posgen = qra_positions.PositionGenerator(loc_distributions, excl_radius, rand_seed)
    locations = posgen.locs

    # Note - these are the same conditions for initializing the Combustion object as is in phys.Flame
    chem = _therm.Combustion(_comps.Fluid(species = rel_fluid.species, T = amb_temp, P = amb_pres))

    # Calculate harm/hazard effects at each position for each leak size
    rel_angle_rads = np.radians(rel_angle)
    try:
        flux_dict = effects.calc_thermal_effects(amb_fluid,
                                                 rel_fluid,
                                                 rel_angle=rel_angle_rads,
                                                 site_length=facil_length,
                                                 site_width=facil_width,
                                                 orifices=orifices,
                                                 rel_humid=rel_humid,
                                                 not_nozzle_model=nozzle_model,
                                                 locations=locations,
                                                 developing_flows=developing_flows,
                                                 chem = chem,
                                                 create_plots=create_plots,
                                                 output_dir=output_dir,
                                                 verbose=verbose)
    except ValueError as err:
        if type(rel_species) == dict:
            raise ValueError('Invalid blend provided')

    qrads = flux_dict['fluxes']
    qrad_plot_files = flux_dict['all_pos_files']
    overp_dict = effects.calc_overp_effects(orifices,
                                            nozzle_model,
                                            rel_fluid,
                                            amb_fluid,
                                            rel_angle,
                                            locations,
                                            facil_length,
                                            facil_width,
                                            overp_method,
                                            bst_flame_speed,
                                            tnt_factor,
                                            developing_flows,
                                            create_plots=create_plots,
                                            output_dir=output_dir,
                                            verbose=verbose)
    overpressures = overp_dict['overpressures']
    impulses = overp_dict['impulses']
    overp_plot_files = overp_dict['all_pos_overp_files']
    impulse_plot_files = overp_dict['all_pos_impulse_files']

    if verbose:
        print(f"Heat flux data:\n{qrads}\n Overpressure data:\n{overpressures}\n Impulse data:\n{impulses}")

    # Estimate fatality probabilities
    physical_responses = {'qrads': qrads,
                          'overpressures': overpressures,
                          'impulses': impulses}
    consequence_modeling_decisions = {'probit_thermal_id': probit_thermal_id,
                                      'exposure_time': exposure_time,
                                      'probit_overp_id': probit_overp_id}
    event_consequences = []
    for end_state_name in end_state_consequence_types:
        event_consequence = consequence.calculate_event_consequence(end_state_consequence_types[end_state_name],
                                                                    num_leak_sizes,
                                                                    total_occupants,
                                                                    physical_responses,
                                                                    consequence_modeling_decisions,
                                                                    verbose)
        event_consequences.append(event_consequence)

    # //////////////////////
    # CALCULATE RISK METRICS

    # Risk for each scenario
    if verbose:
        print("Calculating frequencies and consequences for each scenario...")
    scenario_freqs = []
    scenario_fatalities = []
    for i in range(num_leak_sizes):
        # Compute expected events per year for this leak size
        for end_state_probability, consequence_outcome in zip(end_state_probabilities_per_leak[i], event_consequences):
            scenario_freqs.append(end_state_probability * total_leak_freqs[i])
            scenario_fatalities.append(consequence_outcome[i])

    # Overall risk metrics
    if verbose:
        print("Calculating risk for each scenario...")
    plls = risk.calc_all_plls(scenario_freqs, scenario_fatalities)
    if verbose:
        print("Calculating overall risk metrics and risk contributions for each scenario...")
    total_pll, pll_contributions = risk.calc_risk_contributions(plls)
    far = risk.calc_far(total_pll, total_occupants)
    air = risk.calc_air(far, occupant_avg_hours)

    # Collect results for each leak size
    if verbose:
        print("Results for each leak size:")
    end_state_keys = {'Shutdown': 'shut',
                      'No Ignition': 'noig',
                      'Jet Fire': 'jetf',
                      'Explosion': 'expl'}
    for i, leak_result in enumerate(leak_results):
        for j, end_state_name in enumerate(end_state_consequence_types):
            leak_result.list_event_names.append(end_state_name)
            leak_result.list_event_keys.append(end_state_keys[end_state_name])
            leak_result.list_p_events.append(np.around(end_state_probabilities_per_leak[i][j], 20))
            leak_result.list_avg_events.append(np.around(scenario_freqs[i*len(end_state_consequence_types)+j], 20))
            leak_result.list_pll_contrib.append(np.around(pll_contributions[i*len(end_state_consequence_types)+j], 20))
            # event data in dict format for easier GUI consumption
            leak_result.event_dicts = leak_result.get_result_dicts()

        leak_result.mass_flow_rate = discharge_rates[i]
        leak_result.leak_diam = orifices[i].d
        if verbose:
            print(str(leak_result))

    # Re-shape harm values into position table
    position_qrads_reshape = (qrads.reshape((num_leak_sizes, total_occupants))).T
    position_overps_reshape = (overpressures.reshape((num_leak_sizes, total_occupants))).T
    position_impulses_reshape = (impulses.reshape((num_leak_sizes, total_occupants))).T

    results = {
        'total_pll': total_pll,
        'far': far,
        'air': air,
        'leak_results': leak_results,
        'qrad_plot_files': qrad_plot_files,
        'overp_plot_files': overp_plot_files,
        'impulse_plot_files': impulse_plot_files,
        'positions': np.array(locations).transpose(),
        'position_qrads': position_qrads_reshape,
        'position_overps': position_overps_reshape,
        'position_impulses': position_impulses_reshape
    }

    if verbose:
        print(f"PLL: {total_pll:.5E}\nFAR: {far:.5E}\nAIR: {air:.5E}")

        # Print one result key/val pair per line
        print("\nANALYSIS RESULTS:\n{}".format("\n".join(["{}: {}".format(key, val) for key, val in results.items()])))
        print("=== ANALYSIS COMPLETE ===")

    return results



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""


def check_probability_value(probability):
    """
    Checks for valid probability value between 0.0 and 1.0
    """
    if probability < 0:
        raise ValueError(f"Probability value is <0 (value: {probability})")
    if probability > 1:
        raise ValueError(f"Probability value is >1 (value: {probability})")


def calc_probability_not_occur(probability):
    """
    Calculates complementary probability of a given event probability
    """
    probability_not_occur = 1 - probability
    return probability_not_occur


def calc_end_state_probabilities(event_probabilities):
    """
    Calculates the end-state probabilities for an event tree
    given a list of probabilities of "success" of the intermediate events.
    This formulation assumes a "cascading" event tree
    in which the "success" of an intermediate event will lead to an end-state,
    and that "failure" of an intermediate event will lead to the next intermediate event.
    The "failure" of the final intermediate event will lead to the final end-state.
    """
    end_state_probabilities = []
    previous_event_probability = 1.0
    for event_probability in event_probabilities:
        check_probability_value(event_probability)
        end_state_probability = event_probability * previous_event_probability
        end_state_probabilities.append(end_state_probability)
        previous_event_probability *= calc_probability_not_occur(event_probability)
    final_end_state_probability = previous_event_probability
    end_state_probabilities.append(final_end_state_probability)
    return end_state_probabilities


def calc_end_state_frequencies(initiating_event_frequency, end_state_probabilities):
    """
    Calculates the frequency of event tree end-states,
    given the frequency of the initiating event and the end-state probabilities
    """
    end_state_frequencies = []
    for end_state_probability in end_state_probabilities:
        end_state_frequency = initiating_event_frequency * end_state_probability
        end_state_frequencies.append(end_state_frequency)
    return end_state_frequencies



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import os
import warnings

import numpy as np
from scipy import constants as const
from scipy import integrate, interpolate, optimize

from ._jet import DevelopingFlow
from ._therm import Combustion
from ._comps import Fluid
from ._plots import plot_sliced_contour, plot_contour
from ._utils import get_distance_to_effect
from ..utilities.custom_warnings import PhysicsWarning


class Flame:
    def __init__(self, fluid, orifice, ambient, mdot=None,
                 theta0=0, x0=0, y0=0,
                 nn_conserve_momentum=True, nn_T='solve_energy',
                 chem=None,
                 lamf=1.24, lamv=1.24, betaA=3.42e-2, alpha_buoy=5.75e-4,
                 T_establish_min=-1, verbose=False,
                 Smax=np.inf, dS=None, tol=1e-6, 
                 numB=5, n_pts_integral=100, 
                 wind_speed = 0, developing_flow=None):
        '''
        class for calculating the characteristics of a 2-D flame, without wind
        see Ekoto et al. International Journal of Hydrogen Energy, 39, 2014 (20570-20577)
        
        Parameters
        ----------
        fluid: Fluid object (hc_comps)
            fluid that is being released
        orifice: Orifice object (h2_comps)
            orifice through which fluid is being released
        ambient: Fluid object (hc_comps)
            fluid into which release is occurring
        mdot: float, optional 
            should only be specified for subsonic release, mass flow rate (kg/s)
        theta0 : float, optional
            angle of release (rad) default value of 0 is horizontal
        x0 : float, optional
            horizontal starting point (m)
        y0 : float, optional
            vertical starting point (m)
        nn_conserve_momentum: boolean, optional
            whether notional nozzle model should conserve mass and momentum, or mass only,
            together with nn_T determines which notional nozzle model to use (see below)
        nn_T: string, optional
            either 'solve_energy', 'Tthroat' or specified temperature (stagnation temperature) 
            with nn_conserve_momentum leads to one of the following notional nozzle models:
            YuceilOtugen - conserve_momentum = True, T = 'solve_energy'
            EwanMoodie - conserve_momentum = False, T = 'Tthroat'
            Birch - conserve_momentum = False, T = T0
            Birch2 - conserve_momentum = True, T = T0
            Molkov - conserve_momentum = False, T = 'solve_energy'
        chem : chemistry class (see hc_therm for usage), optional
            if none given, will initialize new chemistry class
        lamf : float
            spreading ratio for mixture fraction Gaussian profile
        lamv : float
            spreading ratio for velocity Gaussian profile
        betaA : float
            momentum entrainment coefficient
        alpha_buoy : float
            buoyancy entrainment coefficient    
        T_establish_min: float, optional
            minimum temperature for start of integral model
        verbose : bool, optional
            If True, extra output will be printed (default True)
        Smax: float, optional
            limit of integration, integrator will stop when it reaches minimum of Flame.length() or Smax
        dS: float, optional
            integrator step size, if None, defaults to Smax/Flame.length() solver adds steps in high gradient areas
        tol: float, optional
            relative and absolute tolerance for integrator
        max_steps: float, optional
            maximum steps for integrator
        numB: float, optional
            maximum number of halfwidths (B) considered to be infinity - for integration in equations
        n_pts_integral: int, optional
            maximum number of points in integration (from 0 to numB)
        wind_speed: int, optional
            ambient wind speed (m/s)
        developing_flow: DevelopingFlow object, optional
            speeds up calculation by only initialzing the object once
        '''
        self.x, self.y, self.S = [], [], []
        if developing_flow is None:
            self.developing_flow = DevelopingFlow(fluid, orifice, ambient, mdot,
                                                  theta0=theta0, x0=x0, y0=y0,
                                                  lam=lamf, betaA=betaA,
                                                  nn_conserve_momentum=nn_conserve_momentum, nn_T=nn_T,
                                                  T_establish_min=T_establish_min,
                                                  verbose=verbose)
        else:
            self.developing_flow = developing_flow
            
        self.initial_node = self.developing_flow.initial_node
        self.mass_flow_rate = self.developing_flow.mass_flow_rate
        expanded_plug_node = self.developing_flow.expanded_plug_node

        self.fluid, self.ambient = fluid, ambient
        self.Emom = betaA * np.sqrt(const.pi / 4 * expanded_plug_node.d ** 2 *
                                    expanded_plug_node.rho * expanded_plug_node.v ** 2 / ambient.rho)
        self.lamf, self.lamv, self.alpha_buoy = lamf, lamv, alpha_buoy
        self.chem = chem
        self.verbose = verbose
        self.wind_speed = wind_speed
        self.solve(Smax, dS, tol, numB, n_pts_integral)
        
    @classmethod
    def from_developed_flow(cls, developing_flow,
                            chem=None, 
                            lamf=1.24, lamv=1.24, betaA=3.42e-2, alpha_buoy=5.75e-4,
                            verbose=False,
                            Smax=np.inf, dS=None, tol=1e-6, 
                            numB=5, n_pts_integral=100, 
                            wind_speed = 0):
        '''
        Initialization of a Flame when the DevelopingFlow calculations have already been made.
        These calculations can be slow for a blend, so if both a Jet and Flame are to be created for the same
        release (e.g. during a QRA), doing the calculations a single time can significantly improve overall
        computation time.
        '''
        if developing_flow.lam != lamf:
            developing_flow.lam = lamf
        if developing_flow.betaA != betaA:
            developing_flow.betaA = betaA
        return cls(developing_flow.fluid, developing_flow.orifice, developing_flow.ambient, 
                   mdot=None, 
                   theta0=0, x0=0, y0=0, 
                   nn_conserve_momentum=True, nn_T='solve_energy', 
                   chem=chem,
                   lamf=lamf, lamv=lamv, betaA=betaA, alpha_buoy=alpha_buoy, 
                   T_establish_min=-1, 
                   verbose=verbose,
                   Smax=Smax, dS=dS, tol=tol, 
                   numB=numB, n_pts_integral=n_pts_integral, 
                   wind_speed = wind_speed, developing_flow = developing_flow)

    def _govEqns(self, S, ind_vars, numB=5, n_pts_integral=100):
        '''
        Governing equations for a flame, written in terms of d/dS of (V_cl, B, theta, f_cl, x, and y).
        
        A matrix soluition to the continuity, x-momentum, y-mometum and mixture fraction equations
        solves for d/dS of the dependent variables V_cl, B, theta, and f_cl.  Numerically integrated
        to infinity = numB * B(S) using numpts discrete points.
        '''
        # break independent variables out of ind_vars
        [V_cl, B, theta, f_cl, x, y] = ind_vars
        
        # needed to integrate to infinity (numB*B):
        r = np.zeros(n_pts_integral)
        r[1:] = np.logspace(-7, np.log10(numB * B), n_pts_integral - 1)

        # mixture fraction and velocity have Gaussian shapes
        f = f_cl * np.exp(-(r / (self.lamf * B)) ** 2)
        V = V_cl * np.exp(-(r / (self.lamv * B)) ** 2)

        # density isn't a nice Gaussian, due to combustion 
        try:
            rho = self.chem.rho_prod(f)
            drhodf = self.chem.drhodf(f)
        except:
            warnings.warn('Clipping f - something has gone wrong.', category=PhysicsWarning)
            f = np.clip(f, 0, 1)
            rho = self.chem.rho_prod(f)
            drhodf = self.chem.drhodf(f)

        rho_int = integrate.trapz(self.ambient.rho - rho, r)

        Ebuoy = Ebuoy = (2 * np.pi * self.alpha_buoy * np.sin(theta) * const.g * rho_int /
                         (B * V_cl * self.developing_flow.fluid_exp.rho))  # m**2/s
        E = self.Emom + Ebuoy

        # right-hand side of governing equations:
        RHS = np.array([self.ambient.rho * E / (2 * const.pi),  # continuity
                        self.wind_speed * self.ambient.rho * E / (2 * const.pi),  # x-momentum
                        integrate.trapz((self.ambient.rho - rho) * const.g * r, r),  # y-momentum
                        0])  # mixture fraction

        zero = np.zeros_like(r)
        dfdS = np.array([zero, 2 * r ** 2 / self.lamf ** 2 / B ** 3 * f, zero, f / f_cl])
        dVdS = np.array([V / V_cl, 2 * r ** 2 / self.lamv ** 2 / B ** 3 * V, zero, zero])
        drhodS = drhodf * dfdS
        dthetadS = np.array([zero, zero, np.ones_like(r), zero])

        # left-hand side of governing equations:
        LHS = np.array([drhodS * V * r + rho * dVdS * r,  # continuity
                        drhodS * V ** 2 * np.cos(theta) * r + 2 * rho * V * dVdS * np.cos(
                            theta) * r + rho * V ** 2 * -np.sin(theta) * dthetadS * r,  # x-momentum
                        drhodS * V ** 2 * np.sin(theta) * r + 2 * rho * V * dVdS * np.sin(
                            theta) * r + rho * V ** 2 * np.cos(theta) * dthetadS * r,  # y-momentum
                        drhodS * V * f * r + rho * dVdS * f * r + rho * V * dfdS * r])  # mixture fraction
        LHS = integrate.trapz(LHS, r)
        
        dz = np.append(np.linalg.solve(LHS, RHS), np.array([np.cos(theta), np.sin(theta)]), axis=0)
        return dz

    def solve(self, Smax=np.inf, dS=None, tol=1e-6,
              numB=5, n_pts_integral=100):
        '''
        Solves for a flame. Returns a dictionary of flame results.  Also updates the Flame class with those results.
        
        Parameters
        ----------
        Smax : float, optional
            endopoint along curved flame for integration (m) default will calculate visible length of flame
        
        Returns
        -------
        res : dict
            dictionary of flame results
        '''
        #ESH note: self.developing_flow.fluid_exp is at a much lower temperature than ambient and gives funky heat flux numbers if used in the Combustion object, hence initilization at ambient T and P - could be improved. 
        try:
            if self.chem.Treac != self.ambient.T or abs(self.chem.P / self.ambient.P - 1) > 1e-10:
                self.chem.reinitilize(Fluid(species = self.fluid.species, T = self.ambient.T, P = self.ambient.P))
        except:
            self.chem = Combustion(Fluid(species = self.fluid.species, T = self.ambient.T, P = self.ambient.P))

        if self.verbose:
            print('solving for the flame...', end='')
        Smax = min(Smax, self.length())

        Y_cl0 = self.initial_node.Y_cl
        f_cl0 = optimize.newton(lambda f: Y_cl0 - self.chem._Yreac(f)[self.chem.reac],
                                Y_cl0)  # ESH-04/05/21 - updated _Yprod in line above to _Yreac

        if dS == None:
            max_step = np.inf
            first_step = None
        else:
            max_step = dS
            first_step = dS
        sol = integrate.solve_ivp(self._govEqns, [self.initial_node.S, Smax], 
                                  np.array([self.initial_node.v_cl, self.initial_node.B, self.initial_node.theta, f_cl0, 
                                            self.initial_node.x, self.initial_node.y]),
                                  max_step = max_step, first_step = first_step,
                                  args = (numB, n_pts_integral),
                                  atol=tol, rtol=tol,
                                  method = 'LSODA'
                                  )

        result = dict(zip(['V_cl', 'B', 'theta', 'f_cl', 'x', 'y'], sol.y))
        result['S'] = sol.t
        for k, v in result.items():
            self.__dict__[k] = v
        if self.verbose:
            print('done.')
        return result

    def length(self):
        '''
        These correlations come from Schefer et al. IJHE 31 (2006): 1332-1340
        and Molina et al PCI 31 (2007): 2565-2572
        
        returns the visible flame length
        also updates the flame class with 
        
        .Lvis (flame length), 
        .Wf (flame width), 
        .tauf (flame residence time)
        .Xrad (radiant fraction)
        '''
        try:
            if self.chem.Treac != self.ambient.T or abs(self.chem.P / self.ambient.P - 1) > 1e-10:
                self.chem.reinitilize(Fluid(species = self.fluid.species, T = self.ambient.T, P = self.ambient.P))
        except:
            self.chem = Combustion(Fluid(species = self.fluid.species, T = self.ambient.T, P = self.ambient.P))
        fs, Tad = self.chem.fstoich, self.chem.T_prod(self.chem.fstoich)
        Tamb = self.ambient.T
        rhoair, rhof = self.ambient.rho, self.chem.rho_prod(self.chem.fstoich)
        orifice1, gas1 = self.developing_flow.orifice_exp, self.developing_flow.fluid_exp
        Deff, rhoeff = orifice1.d, gas1.rho

        # Compute the flame Froude number
        Frf = (gas1.v * fs ** 1.5) / (((rhoeff / rhoair) ** 0.25) * np.sqrt(((Tad - Tamb) / Tamb) * const.g * Deff))
        self.Frf = Frf

        # Compute visible flame length
        Lstar = ((13.5 * Frf ** 0.4) / (1 + 0.07 * Frf ** 2) ** 0.2) * (Frf < 5) + 23 * (Frf >= 5)

        dstar = Deff * (rhoeff / rhoair) ** 0.5
        self.dstar = dstar

        self.Lvis = Lstar * dstar / fs  # visible flame length [m]
        self.Wf = 0.17 * self.Lvis
        # flame residence time [ms]

        self.tauf = (const.pi / 12) * (rhof * (self.Wf ** 2) * self.Lvis * fs) / (orifice1.mdot(gas1)) * 1000
        self.Xrad = 9.45e-9 * (self.tauf * self.chem.absorption_coeff * Tad ** 4) ** 0.47  # see Panda, Hecht, IJHE 2016
        mdot = self.developing_flow.orifice_exp.mdot(self.developing_flow.fluid_exp)  # mass flow rate [kg/s]
        self.Srad = self.Xrad * mdot * self.chem.DHc

        return self.Lvis

    def Qrad_multi(self, x, y, z, RH, WaistLoc=0.75, N=50):
        '''
        MultiSource radiation model
        follows Hankinson & Lowesmith, CNF 159, 2012: 1165-1177       
        '''
        obsOrg = np.array([x, y, z]).T

        try:
            Lvis = self.Lvis  # length of visible flame [m]
        except:
            Lvis = self.length()
        T = self.ambient.T

        n = int(WaistLoc * N)
        w = np.arange(1, N + 1, dtype=float)
        w[n:] = (n - ((n - 1) / (N - (n + 1))) * (w[n:] - (n + 1)))
        w /= np.sum(w)

        try:
            S = np.linspace(self.S[0], min([self.S[-1], self.Lvis]), N)
            X = interpolate.interp1d(self.S, self.x)(S)
            Y = interpolate.interp1d(self.S, self.y)(S)
        except:
            warnings.warn('Running flame model with default parameters.', category=PhysicsWarning)
            self.solve()
            S = np.linspace(self.S[0], min([self.S[-1], self.Lvis]), N)
            X = interpolate.interp1d(self.S, self.x)(S)
            Y = interpolate.interp1d(self.S, self.y)(S)

        sourceOrg = np.array([X, Y, np.zeros_like(X)]).T

        Qrad = np.zeros(obsOrg.shape[:-1])

        for j in range(len(w)):
            v = sourceOrg[j] - obsOrg
            if obsOrg.ndim > 1:
                obsNorm = v / (np.linalg.norm(v, axis=1) + 1e-99)[:, np.newaxis]
            else:
                obsNorm = v / (np.linalg.norm(v) + 1e-99)

            if obsOrg.ndim == 4:
                len_v = np.linalg.norm(v, axis=3)
                phi = np.arcsin(np.linalg.norm(np.cross(obsNorm, v), axis=3) / len_v)
            elif obsOrg.ndim == 3:
                len_v = np.linalg.norm(v, axis=2)
                phi = np.arcsin(np.linalg.norm(np.cross(obsNorm, v), axis=2) / len_v)
            elif obsOrg.ndim == 2:
                len_v = np.linalg.norm(v, axis=1)
                phi = np.arcsin(np.linalg.norm(np.cross(obsNorm, v), axis=1) / len_v)
            elif obsOrg.ndim == 1:
                len_v = np.linalg.norm(v)
                phi = np.arcsin(np.linalg.norm(np.cross(obsNorm, v)) / len_v)
            Qrad += w[j] / (4 * const.pi * len_v ** 2) * np.cos(phi) * self.Srad * calc_transmissivity(len_v, T, RH)

        return Qrad.T

    @property
    def _contourdata(self):
        iS = np.arange(len(self.S))
        r = np.append(
            np.append(-np.logspace(np.log10(10 * max(self.B)), -2), np.linspace(-10 ** -2.1, 10 ** -2.1, num=10)),
            np.logspace(-2, np.log10(10 * max(self.B))))
        r, iS = np.meshgrid(r, iS)
        B = self.B[iS]
        f_cl = self.f_cl[iS]

        fvals = f_cl * np.exp(-(r / (self.lamf * B)) ** 2)
        Tvals = self.chem.T_prod(fvals)

        x = self.x[iS] + r * np.sin(self.theta[iS])
        y = self.y[iS] - r * np.cos(self.theta[iS])
        return x, y, Tvals

    def plot_temps(self, mcolors='w', xlims=None, ylims=None, vlims=None,
                   contour_levels=None, addColorBar=True, plot_color=None,
                   plot_title=None, x_label=None, y_label=None,
                   fig_params={}, plot_params={}, subplot_params={}, fig=None, ax=None):
        '''
        Creates temperature contour plots.
        Sends parameters to phys._plots plot_contour function and runs it.
        '''
        plot_contour(data_type='temperature', jet_or_flame=self,
                     mcolors=mcolors, xlims=xlims, ylims=ylims, vlims=vlims,
                     contour_levels=contour_levels, add_colorbar=addColorBar, plot_color=plot_color,
                     plot_title=plot_title, x_label=x_label, y_label=y_label,
                     fig_params=fig_params, plot_params=plot_params, subplot_params=subplot_params, fig=fig, ax=ax)
        
    def plot_heat_flux_sliced(self, title=None,
                              filename='HeatFluxSliced.png',
                              directory=os.getcwd(),
                              RH=0.89,
                              contours=None,
                              nx=50, ny=50, nz=50,
                              xlims=None, ylims=None, zlims=None,
                              WaistLoc=0.75,
                              savefig=True):
        '''
        plots slices of heat flux levels

        Parameters
        ----------
        title : string (optional)
            title shown on plot
        filename : string, optional
            file name to write
        directory : string, optional
            directory in which to save file
        RH : float
            relative humidity
        contours : ndarray or list (optional)
            contour levels shown on plot in kW/m^2
            (default values are 2012 International Fire Code (IFC) exposure limits
            for property lines (1.577 kW/m2),
            employees (4.732 kW/m2),
            and non-combustible equipment (25.237 kW/m2))
        nx, ny, nz: float (optional)
            number of points to solve for the heat flux in the x, y, and z directions
        xlims, ylims, zlims : tuple (optional)
            plot limits of (min, max) in each dimension 
        WaistLoc : float (optoinal)
            value between 0 and 1 along flame length at which to make xz slice
        savefig : boolean (optional)
            whether to save the figure as filename

        Returns
        -------
        If savefig is True, returns full filepath of saved figure;
        if savefig is False, returns figure object.
        '''
        if contours is None:
            contours = [1.577, 4.732, 25.237]  # kW/m2

        Lvis = self.length()
        flameCen = [np.interp(Lvis * WaistLoc, self.S, self.x),
                    np.interp(Lvis * WaistLoc, self.S, self.y),
                    0]

        flame_centerlines = [self.x[self.S <= self.Lvis],
                             self.y[self.S <= self.Lvis],
                             np.ones_like(self.y[self.S <= self.Lvis]) * flameCen[2]]

        colorbar_label = 'Heat Flux [kW/m$^2$]'

        fig_or_filepath = plot_sliced_contour(contours, xlims, ylims, zlims, nx, ny, nz,
                                              self.calc_distance_to_heatflux, self.Qrad_multi,
                                              flameCen, colorbar_label,
                                              origin_lines=flame_centerlines,
                                              title=title, savefig=savefig,
                                              directory=directory, filename=filename,
                                              RH=RH, WaistLoc=WaistLoc)
        return fig_or_filepath

    def generate_positional_flux(self, flux_coordinates, rel_humid):
        """ Calculate flux at positions according to radiative source model

        Parameters
        ----------
        flux_coordinates : list of locations
            List of locations at which to determine flux,
            each location is a tuple of 3 coordinates (m):
            [(x1, y1, z1), (x2, y2, z2), ...]

        rel_humid : flat
            relative humitidy

        Returns
        -------
        flux : ndarray
            flux values at specified positions (W/m^2)
        """
        
        x_values = [flux_coordinate[0] for flux_coordinate in flux_coordinates]
        y_values = [flux_coordinate[1] for flux_coordinate in flux_coordinates]
        z_values = [flux_coordinate[2] for flux_coordinate in flux_coordinates]
        flux = self.Qrad_multi(x_values, y_values, z_values, rel_humid)
        return flux

    def get_srad(self):
        """
        Returns : float
            Total emitted radiative power for flame (W)
        """
        return self.Srad

    def get_visible_length(self):
        """
        Returns : float
            Visible flame length (m)
        """
        return self.Lvis
    
    @property    
    def birds_eye_flame_length(self):
        '''
        returns : float 
            flame length from above
        '''
        return np.interp(self.Lvis, self.S, self.x)

    def calc_distance_to_heatflux(self, heat_flux_level, direction='x',
                                  RH=0.89, WaistLoc=0.75,
                                  max_distance=500, negative_direction=False):
        '''
        Calculate distance from leak point to a given heatflux
        in the direction specified

        Parameters
        ----------
        heat_flux_level : float
            heat flux level for which to get the distance to (W/m^2)
        direction : 'x', 'y', or 'z', optional
            direction for which to calculate the distance
            (default is 'x')
        RH : float
            relative humidity
        WaistLoc : float  
            fractional distance along flame at which to look for the distance
        max_distance : float (optional)
            maximum distance to look for heat flux level
        negative_direction : Boolean (optional)
            whether or not to look in the negative direction instead of positive
            (default is False)

        Returns
        -------
        distance : float
            distance to heat_flux_level (m)
        '''
        flame_center_streamline = self.Lvis * WaistLoc
        flame_center = [
            np.interp(flame_center_streamline, self.S, self.x),  # x
            np.interp(flame_center_streamline, self.S, self.y),  # y
            0  # z
        ]
        from_point = flame_center
        if direction == 'x':
            from_point[0] = 0
        elif direction == 'y':
            from_point[1] = 0
        elif direction == 'z':
            from_point[2] = 0
        else:
            raise ValueError(f"Direction ('{direction}') must be 'x', 'y', or 'z'")
        distance = get_distance_to_effect(value=heat_flux_level,
                                          from_point=from_point,
                                          direction=direction,
                                          effect_func=self.Qrad_multi,
                                          max_distance=max_distance,
                                          negative_direction=negative_direction,
                                          RH=RH)
        return distance


def calc_transmissivity(path_length, ambient_temperature, relative_humidity, atmospheric_CO2_ppm=335):
    '''
    Calculates atmospheric transmissivity from:
    Wayne, J. Loss Prev. Proc. Ind. 1991

    Parameters
    ----------
    path_length : float
        Path length (m) through which radiative light must travel

    ambient_temperature : float
        Ambient temperature (K)

    relative_humidity : float
        Fractional relative humidity (0-1)

    atmospheric_CO2_ppm : float
        Atmospheric CO2 concentration (ppm),
        default is 335 ppm
    
    Returns
    -------
    transmissivity : float
        Atmospheric transmissivity 
    '''
    sat_water_vap_pressure_mmHg = np.exp(20.386 - 5132 / ambient_temperature)  # mmHg, saturated vapor pressure from Wikipedia
    XH2O = relative_humidity * path_length * sat_water_vap_pressure_mmHg * 2.88651e2 / ambient_temperature
    XCO2 = path_length * 273. / ambient_temperature * atmospheric_CO2_ppm / 335.
    transmissivity = (1.006 - 0.01171 * np.log10(XH2O) - 0.02368 * np.log10(XH2O) ** 2
                       - 0.03188 * np.log10(XCO2) + 0.001164 * np.log10(XCO2) ** 2)
    return transmissivity



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""


def get_default_leak_sizes():
    default_leak_sizes = [0.01, 0.10, 1, 10, 100]
    return default_leak_sizes



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

# TODO: add tests for each of the specific position distribution types

import unittest

from hyram.qra import positions


class TestPositionGenerator(unittest.TestCase):
    """
    Tests of generation of positions for QRA
    """
    def setUp(self):
        loc_distributions = [
            [1,  # number of occupants for this distribution
             ('deterministic', 1, None),  # distribution and parameters for x-direction
             ('deterministic', 2, None),  # distribution and parameters for y-direction
             ('deterministic', 3, None)],  # distribution and parameters for z-direction
            [2,  # number of occupants for this distribution
             ('uniform', 4, 5),  # distribution and parameters for x-direction
             ('deterministic', 6, None),  # distribution and parameters for y-direction
             ('deterministic', 7, None)]  # distribution and parameters for z-direction
        ]
        self.excl_radius = 0.01  # meters
        self.rand_seed = 3632850
        posgen = positions.PositionGenerator(loc_distributions, self.excl_radius, self.rand_seed)
        self.posgen = posgen

    def test_get_locations(self):
        num_positions = len(self.posgen.locs)
        self.assertEqual(num_positions, 3)

    def test_get_xlocs(self):
        xlocs = self.posgen.get_xlocs()
        self.assertEqual(len(xlocs), 3)

    def test_get_ylocs(self):
        ylocs = self.posgen.get_ylocs()
        self.assertListEqual(ylocs, [2, 6, 6])

    def test_get_zlocs(self):
        zlocs = self.posgen.get_zlocs()
        self.assertListEqual(zlocs, [3, 7, 7])

    def test_zero_occupants(self):
        loc_distributions = [
            [0,  # number of occupants for this distribution
             ('deterministic', 1, None),  # distribution and parameters for x-direction
             ('deterministic', 2, None),  # distribution and parameters for y-direction
             ('deterministic', 3, None)],  # distribution and parameters for z-direction
        ]
        posgen = positions.PositionGenerator(loc_distributions, self.excl_radius, self.rand_seed)
        locations = posgen.locs
        self.assertListEqual(locations, [])

    def test_reject_too_many_dists(self):
        loc_distributions = [
            [1,  # number of occupants for this distribution
             ('deterministic', 1, None),  # distribution and parameters for x-direction
                                            # y-direction missing
             ('deterministic', 3, None)],  # distribution and parameters for z-direction
        ]
        self.assertRaises(ValueError, positions.PositionGenerator, loc_distributions)

    def test_reject_too_few_dists(self):
        loc_distributions = [
            [1,  # number of occupants for this distribution
             ('deterministic', 1, None),  # distribution and parameters for x-direction
             ('deterministic', 2, None),  # distribution and parameters for y-direction
             ('deterministic', 2, None),  # extra distribution
             ('deterministic', 3, None)],  # distribution and parameters for z-direction
        ]
        self.assertRaises(ValueError, positions.PositionGenerator, loc_distributions)

    def test_reject_negative_number_occupants(self):
        loc_distributions = [
            [-1,  # negative number of occupants for this distribution
             ('deterministic', 1, None),  # distribution and parameters for x-direction
             ('deterministic', 2, None),  # distribution and parameters for y-direction
             ('deterministic', 3, None)],  # distribution and parameters for z-direction
        ]
        self.assertRaises(ValueError, positions.PositionGenerator, loc_distributions)

    def test_reject_non_integer_number_occupants(self):
        loc_distributions = [
            [1.1,  # float number of occupants for this distribution
             ('deterministic', 1, None),  # distribution and parameters for x-direction
             ('deterministic', 2, None),  # distribution and parameters for y-direction
             ('deterministic', 3, None)],  # distribution and parameters for z-direction
        ]
        self.assertRaises(ValueError, positions.PositionGenerator, loc_distributions)

    def test_reject_location_inside_exclusion_radius(self):
        loc_distributions = [
            [1,  # float number of occupants for this distribution
             ('deterministic', 0.005, None),  # distribution and parameters for x-direction
             ('deterministic', 0, None),  # distribution and parameters for y-direction
             ('deterministic', 0, None)],  # distribution and parameters for z-direction
        ]
        self.assertRaises(ValueError, positions.PositionGenerator, loc_distributions, self.excl_radius)



if __name__ == "__main__":
    unittest.main()



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""


class PhysicsWarning(UserWarning):
    """
    Generic warning class to bubble up pertinent warnings to GUI

    Attributes
    ----------
    message : str
        message describing reason for the warning
    """

    def __init__(self, message):
        self.message = message





"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import warnings

from CoolProp import CoolProp
import numpy as np
from scipy import optimize, interpolate
from scipy import constants as const

from ._fuel_props import FuelProperties
from ..utilities.custom_warnings import PhysicsWarning



class CoolPropWrapper:
    def __init__(self, species):
        '''
        Class that uses CoolProp for equation of state calculations.
        
        phase: 'gas' or 'liquid' for fluid at saturated vapor pressure
        '''
        self._cp = CoolProp
        self.spec = species
        self.MW = self._cp.PropsSI(self.spec, 'molemass')        
    
    def _init_fluid_params(self, params):
        '''
        Initializes Fluid properties.
        Calculates 2 remaining properties given 2 of rho, pressure, temperature, or phase
        
        Parameters
        ----------
        params : dict
            Dict of Fluid parameters keyed by the labels 'D', 'P', 'T', and 'phase'
            Valid input requires 2 of the parameters to be defined and 2 to be None

        Returns
        -------
        params : dict
            Returns the same dict with all values filled in
        '''
        # Set input and output values
        output = [key for key, val in params.items() if val is None]
        input_dict = {key:val for key, val in params.items() if val is not None}

        if len(output) != 2 or len(input_dict) != 2:
            raise ValueError("Fluid not properly defined - too many or too few fluid initialization variables")

        # Put keys and values in seperate tuples
        input_params, input_values = zip(*input_dict.items())

        if 'phase' in input_params:
            self.phase = input_dict.get('phase')
            # Convert phase string to Q value
            input_dict.pop('phase')
            input_dict.update({'Q':{'gas':1, 'liquid':0}[self.phase]})
            props = self.PropsSI(output, **input_dict)
        else:
            # Change phase output to Q for CoolProp to handle
            idx = output.index('phase')
            output[idx] = 'Q'

            # Handle special case for calculating density at saturation
            if 'D' in output and 'Q' in output:
                try: 
                    props = self.PropsSI(output, **input_dict)
                except ValueError:
                    print('exception')
                    warnings.warn('Assuming gas phase to calculate density.', category=PhysicsWarning)
                    self.phase = 'gas'
                    rho = self._cp.PropsSI('D', input_params[0]+self.phase, input_values[0], input_params[1], input_values[1], self.spec)
                    props = [rho, self.phase]
            else:
                props = self.PropsSI(output, **input_dict)

            try:
                if '&' in self.spec: 
                    raise ValueError
                self.phase = self._cp.PhaseSI(input_params[0], input_values[0], input_params[1], input_values[1], self.spec)
            except ValueError:
                Q = props[idx]
                if (Q < 1) and (Q > 0):
                    self.phase = 'twophase'
                elif Q == 1:
                    self.phase = 'vapor'
                elif Q == 0:
                    self.phase = 'liquid'
                else:
                    self.phase = ''

        # Put calculated values into the params dict
        params[output[0]] = props[0]
        params[output[1]] = props[1]

        # Overwrite phase info with string
        params['phase'] = self.phase
        return params
    
    def get_property(self, output, **kwargs):
        '''
        General method for calculating one or more fluid properties

        Parameters
        ----------
        output : string or list of strings
            Same output parameters accepted by CoolProp.PropsSI (e.g., T, P, S, D) with additional custom keywords
                - 'H0' calculates total enthalpy using 'P', 'D', and 'v'

        keyword args : input parameters
            Pair of known parameters entered as individual params (e.g., T=273, P=1e6, etc.)
            An extra 3rd parameter can be added to define either 'phase' or 'v'

        Returns
        -------
        props : float or list of floats
            Return values based on the 'output' parameter

        Examples
        --------
        total_enthalpy = get_property('H0', P=fluid.P, D=rho, v=v)
        rho = get_property('D', T=T, P=P, phase=liquid)
        rho = get_property('D', T=T, P=P)
        s = get_property('S', D=fluid.rho, T = fluid.T)
        h, rho = get_property(['H', 'D'], P=P, S=s0)
        '''
        input_dict = {key:val for key, val in kwargs.items() if val is not None}

        # H0 keyword
        if output == 'H0':
            try:
                v = input_dict.pop('v')
                h = self.PropsSI('H', **input_dict)
            except Exception as err:
                if self.phase == '':
                    raise ValueError(f'Unable to find enthalpy of fluid with {input_dict}')
                warnings.warn(f'Recalculating with specification of phase {self.phase}.')
                h = self.PropsSI('H', phase=self.phase, **input_dict)
            return h + v**2/2

        # General case
        try:
            out = self.PropsSI(output, **input_dict)
        except:
            raise ValueError(f'Unable to calculate {output} using input parameters {input_dict}')
        return out

    def PropsSI(self, output, **kwargs):
        '''
        Wrapper on CoolProps PropsSI
        
        Parameters 
        ----------
        output : list of strings
            Same parameters accepted by CoolProp.PropsSI (e.g., T, P, S, D)
        
        Returns
        -------
        Outputs from CoolProp listed within output (could be single value or list)
        '''
        if 'phase' in kwargs:
            phase =  '|' + kwargs.pop('phase')
        else:
            phase = ''        
        try:
            if '&' in self.spec:
                raise ValueError

            (k1, v1), (k2, v2) = kwargs.items()
            k1 += phase
            if type(v1) == list or type(v1) == np.ndarray: 
                if len(v1) == 1:
                    v1 = float(v1)
            if type(v2) == list or type(v2) == np.ndarray:
                if len(v2) == 1:
                    v2 = float(v2)

            try:
                out = self._cp.PropsSI(output, k1, v1, k2, v2, self.spec)
            except:
                warnings.warn('PropsSI could not find a solution - estimating', category=PhysicsWarning)

                # Estimate using the first input property
                for i in range(0, 5000):
                    try:
                        result_high = self._cp.PropsSI(output, k1, v1+i, k2, v2, self.spec)
                        result_low = self._cp.PropsSI(output, k1, v1-i, k2, v2, self.spec)
                        break
                    except:
                        continue

                if len(output) == 1:
                    out = (result_high + result_low) / 2
                else:
                    out = []
                    for idx in range(len(output)):
                        high = result_high[idx]
                        low = result_low[idx]
                        out.append((high + low) / 2)
                    
            return out

        except ValueError:  # likely blend
            if '&' not in self.spec:
                return ValueError  # not a blend - some other issue

            spec_names = '&'.join([s.split('[')[0] for s in self.spec.split('&')])
            molefracs = [float(s.split('[')[1][:-1]) for s in self.spec.split('&')]
            eos = self._cp.AbstractState('HEOS', spec_names)
            eos.set_mole_fractions(molefracs)
            guess = self._cp.PyGuessesStructure()

            def out(k):
                if   k == 'D': return eos.rhomass()
                elif k == 'S': return eos.smass()
                elif k == 'H': return eos.hmass()
                elif k == 'U': return eos.umass()
            if ('P' in kwargs) and ('T' in kwargs):
                P = kwargs['P']; T = kwargs['T']
                try:
                    eos.update(self._cp.PT_INPUTS, P, T)
                except:
                    for i in range(10):
                        guess.rhomolar = (i+1)*2*P/(8.314*T) # unclear what the 'right' guess is (hence the loop)
                        try:
                            eos.update_with_guesses(self._cp.PT_INPUTS, P, T, guess)
                            break
                        except:
                            pass
            elif ('T' in kwargs):
                T = kwargs.pop('T')
                k, v = list(kwargs.items())[0]
                def err(P):
                    try:
                        eos.update(self._cp.PT_INPUTS, P, T)
                    except:
                        for i in range(10):
                            guess.rhomolar = (i+1)*2*P/(8.314*T) # unclear what the 'right' guess is (hence the loop)
                            try:
                                eos.update_with_guesses(self._cp.PT_INPUTS, P, T, guess)
                                break
                            except:
                                pass
                    return v - out(k)
                P = optimize.root(err, 101325.)['x'] # somewhat arbitrarily starts at 1 atm
            elif ('P' in kwargs):
                P = kwargs.pop('P')
                k, v = list(kwargs.items())[0]
                def err(T):
                    try:
                        eos.update(self._cp.PT_INPUTS, P, T)
                    except:
                        for i in range(10):
                            guess.rhomolar = (2*i+1)*P/(8.314*T) # unclear what the 'right' guess is (hence the loop)
                            try:
                                eos.update_with_guesses(self._cp.PT_INPUTS, P, T, guess)
                                break
                            except:
                                pass

                    return v - out(k)
                T = optimize.brentq(err, eos.Tmin(), eos.Tmax()) # not the most efficient here, but works
            else:
                (k1, v1), (k2, v2) = list(kwargs.items())
                warnings.warn('Only PT_inputs allowed for blends (this may not work) trying to solve for %s, %s inputs' % (k1, k2), 
                              category=PhysicsWarning)
                def err(TP):
                    T, P = TP
                    try:
                        eos.update(self._cp.PT_INPUTS, P, T)
                    except:
                        for i in range(10):
                            guess.rhomolar = (i+1)*2*P/(8.314*T) # unclear what the 'right' guess is (hence the loop)
                            try:
                                eos.update_with_guesses(self._cp.PT_INPUTS, P, T, guess)
                                break
                            except:
                                pass
                    return [v1 - out(k1), v2 - out(k2)]
                T, P = optimize.root(err, [(eos.Tmin() + eos.Tmax())/2, 101325])['x']              

            def map_outputs(key_or_keys, eos):
                def out(k):
                    if   k == 'D': return eos.rhomass()
                    elif k == 'T': return eos.T()
                    elif k == 'P': return eos.p()
                    elif k == 'S': return eos.smass()
                    elif k == 'H': return eos.hmass()
                    elif k == 'U': return eos.umass()
                    elif k == 'Q': return eos.Q()
                    elif k == 'A': return eos.speed_sound()
                    elif k == 'C': return eos.cpmass()
                    elif k == 'V': return eos.viscosity()
                    elif k == 'isobaric_expansion_coefficient': return eos.isobaric_expansion_coefficient()
                if type(key_or_keys) == list:
                    return [out(k) for k in key_or_keys]
                else:
                    return out(key_or_keys)
            return map_outputs(output, eos)
        except:
            raise warnings.warn('system not properly defined')


class Combustion:
    def __init__(self, fluid, #ambient, # TODO: add ambient object
                 numpoints = 100, verbose = False):
        '''
        Class that performs combustion chemistry calculations.
        Stoichiometry: C_nH_(2n) + eta/2 (O_2 + 3.76N_2) -> max(0, 1-eta) C_nH_(2n) + min(1, eta) H_2O + min(n*eta, n)CO_2 + max(0, (eta-1)/2) O2 + 3.76 eta/2 N2
        
        Initilizes some interpolating functions for T_prod, MW_prod, rho_prod, and drhodf
        
        Parameters
        ----------
        fluid : hyram.phys.Fluid object
            fluid being combusted
        ambient: hyram.phys.Fluid object with ambient air
            air with with the fluid is being combusted
        numpoints : int
            number of points to solve for temperature to create 
            interpolating functions, default value is 100
        verbose: boolean
            whether to include some print statements
            
        Contents
        --------
        self.T_prod(f): array_like
            temperature of products (K) at a mixture fraction, f
        self.MW_prod:
            mixture averaged molecular weight of products (g/mol) at a mixture fraction, f 
        '''
        fuel_props = FuelProperties(fluid.species)
        self.xO2stoich = fuel_props.nC + fuel_props.nH/4 - fuel_props.nO/2

        self.Treac, self.Preac, self.rho_reac = fluid.T, fluid.P, fluid.rho
        # TODO: # Tair, P = ambient.T, ambient.P
        if verbose:
            print('initializing chemistry...', end = '')
        self.reac = fluid.species
        self._nC, self._nH, self._nO = fuel_props.nC, fuel_props.nH, fuel_props.nO

        self.DHc = fuel_props.dHc # heat of combustion, J/kg

        self.therm = {spec:CoolPropWrapper(spec) for spec in ['O2', 'N2', 'H2O', 'CO2', self.reac]}
        self.MW = {spec:therm.MW for spec, therm in self.therm.items()}
        self.fstoich = self.MW[self.reac]/(self.MW[self.reac] + self.xO2stoich * (self.MW['O2'] + self.MW['N2']*3.76))
        ifstoich = int(max(numpoints*self.fstoich, 5))
        fvals = np.append(np.linspace(0, self.fstoich, int(max(numpoints*self.fstoich, 5))), 
                          np.linspace(self.fstoich, 1, int(max(numpoints*(1-self.fstoich), 5))))
        T = self._T_combustion(self.Treac, fvals)
        MWvals = self._MWmix(self._Yprod(fvals))
        
        # Creates some interpolating functions - only create them once during initialization to save computational time later
        self.MW_prod = lambda f: self._MWmix(self._Yprod(f))
        self.T_prod = interpolate.interp1d(fvals, T)
        self.rho_prod = lambda f: self.Preac*self.MW_prod(f)/(const.R*self.T_prod(f))
        self.drhodf = interpolate.interp1d(fvals,
                                           self.Preac/(const.R*T)*(np.append(np.gradient(MWvals[:ifstoich], fvals[:ifstoich]),
                                                                   np.gradient(MWvals[ifstoich:], fvals[ifstoich:])) - 
                                                                   np.append(np.gradient(T[:ifstoich], fvals[:ifstoich]),
                                                                   np.gradient(T[ifstoich:], fvals[ifstoich:]))/T*MWvals))
        self.absorption_coeff = self._plank_mean_absorption_coefficient_stoich()
        if verbose:
            print('done.')

    def reinitilize(self, fluid, numpoints = 100):
        '''
        Reinitializes class to new temperature, pressure, etc.  Can be used rather
        than creating a new instance taking up additional memory.'''
        self.__init__(fluid, numpoints)

    def _MWmix(self, Y):
        '''returns the mixture averaged molecular weight, given a mass fraction'''
        MWmix = 0
        for spec, Yval in Y.items():
            MWmix += Yval/self.MW[spec]
        MWmix = 1/MWmix
        return MWmix
   
    def _eta(self, f):
        '''returns eta, given a mixture fraction'''
        MW = self.MW
        eta = 1/self.xO2stoich*(MW[self.reac]/(1.e-99*np.ones_like(f) +f) - MW[self.reac])/(MW['O2'] + 3.76*MW['N2'])
        return eta
    
    def _Yprod(self, f):
        '''
        the mass fractions of combustion products as a function of the mixture fraction
        todo: needs to be fixed for when CO2 (?and N2?) is in the reactants

        Parameters
        ----------
        f = mixture fraction

        Returns
        -------
        Y - dictionary of mass fractions (kg/kg)
        '''
        eta = self._eta(f)
        nCreac, nHreac, nOreac = self._nC, self._nH, self._nO
        Y = {self.reac:(1-eta)*(eta < 1)*self.MW[self.reac],
             'CO2':nCreac*(eta*(eta <= 1) + (eta > 1))*self.MW['CO2'],
             'H2O':nHreac/2*(eta*(eta <= 1) + (eta > 1))*self.MW['H2O'],
             'O2':(eta-1)*self.xO2stoich*(eta > 1)*self.MW['O2'],
             'N2':eta*self.xO2stoich*3.76*self.MW['N2']}
        s = sum(list(Y.values()))
        for k in Y.keys():
            Y[k] /= s
        return Y
    
    def _Yreac(self, f):
        '''
        the mass fractions of combustion reactants as a function of the mixture fraction
        (assumes that there is no H2O or CO2 as a reactant)
        '''
        Y = {self.reac: self.MW[self.reac], 
             'CO2':     0,
             'H2O':     0,
             'O2':      self._eta(f)*self.xO2stoich*self.MW['O2'],
             'N2':      self._eta(f)*self.xO2stoich*3.76*self.MW['N2']}
        s = sum(list(Y.values()))
        for k in Y.keys():
            Y[k] /= s
        return Y
    
    def _H(self, T, Y, Hdict):
        '''enthalpy of a mixture at a given temperature and pressure'''
        H = np.zeros_like(list(Y.values())[0])
        for species, Yval in Y.items():
            H += Hdict[species](T)*Yval
        return H
        
    def _Hdict(self, Tmin, Tmax = 6000, npoints = 500):
        '''returns dictionary of interpolating enthalpy functions from Tmin to Tmax'''
        Hdict = {}
        for spec in self.therm.keys():
            T = np.linspace(np.max([Tmin, self.therm[spec]._cp.PropsSI('T_min', spec)+0.1]), Tmax, npoints)
            try:
                Hdict[spec] = interpolate.interp1d(T, self.therm[spec].get_property('H', T = T, P = self.Preac), 
                                                   fill_value = 'extrapolate')
            except:
                Hdict[spec] = interpolate.interp1d(T, [self.therm[spec].get_property('H', T = Tval, P = self.Preac) for Tval in T], 
                                                   fill_value = 'extrapolate')
        return Hdict

    def _T_combustion(self, T_reac, f, numpoints = 500):
        '''combustion temperature (K)'''
        DHc = self.DHc*self._Yprod(f)['H2O']/(self._nH/2)*self.MW[self.reac]/self.MW['H2O'] # heat of combustion [J/kg_total_products]
        Hdict = self._Hdict(T_reac, npoints = numpoints)
        H0 = self._H(T_reac, self._Yreac(f), Hdict)
        H0 *= self._MWmix(self._Yreac(f))/self._MWmix(self._Yprod(f)) #J/kg
        H = H0 + DHc
        T = optimize.root(lambda T: self._H(T, self._Yprod(f), Hdict) - H, T_reac*np.ones_like(np.array(f)))['x']
        return T
        
    def _mean_absorption_coefficient(self, species, temperature, pressure=const.atm):
        '''
        Calculates mean absorption coefficient m^-1 for a species at a temperature and pressure

        Parameters
        ----------
        species: string
            species of interest (either H2O or CO2)
        temperaure: float
            temperature (K)
        pressure: float
            pressure (Pa)

        Returns
        -------
        mean_absorption_coefficient: float
            mean absorption coefficient (m)^-1
            
        Note: Formulas and data from Chmielewski and Gieras, "Planck Mean Absorption Coefficients of H2O, CO2, CO and NO 
        for radiation numerical modeling in combusting flows," Journal of Power Technogies 95 (2) 2015: 97-104. 
        Calculations agree with https://doi.org/10.1016/S0022-4073(01)00178-9
        '''
        data = {'H2O':[[63.87, 149.5, 174.3],
                       [2.629, 493.6, 111.1],
                       [222.9, -2110, 1592],
                       [0.991, 1700, 619.2]],
                'CO2':[[34.57, 45.32, 838.4],
                       [1.22, 930.2, 184.9],
                       [-22.32, 323.9, 217.9],
                       [17.14, 92.98, 1657]]}
        
        coeffs = data[species]
        ap_mean = sum([coeff[0] * np.exp(-((temperature - coeff[1]) / coeff[2]) ** 2)
                       for coeff in coeffs])
        return ap_mean * pressure / const.bar
        
    def _plank_mean_absorption_coefficient(self, temperature, x_H2O, x_CO2, pressure=const.atm):
        '''
        Calculates Plank mean absorption coefficient m^-1 for a mixture of water vapor and CO2

        Parameters
        ----------
        temperaure: float
            temperature (K)
        x_H2O: float
            mole fraction water
        x_CO2: float
            mole fraction CO2
        pressure: float
            pressure (Pa)

        Returns
        -------
        plank_mean_absorption_coefficient: float
            Plank mean absorption coefficient (m)^-1
        '''
        return sum([x * self._mean_absorption_coefficient(spec, temperature, pressure) 
                    for x, spec in zip([x_H2O, x_CO2], ['H2O', 'CO2'])])
                    
    def _plank_mean_absorption_coefficient_stoich(self):
        '''
        Calculates Plank mean absorption coefficient at stiochiometric combustion conditions - used in radiation calculations
        
        Returns
        -------
        plank_mean_absorption_coefficient: float
            Plank mean absorption coefficient (m)^-1 at stoichiometric conditions
        '''
        Y = self._Yprod(self.fstoich)
        x_H2O = Y['H2O'] / self.MW['H2O'] * self._MWmix(Y)
        x_CO2 = Y['CO2'] / self.MW['CO2'] * self._MWmix(Y)
        T = self.T_prod(self.fstoich)
        return self._plank_mean_absorption_coefficient(T, x_H2O, x_CO2, self.Preac)
        
    def dP_expansion(self, enclosure, mass):
        '''
        Pressure due to the expansion of gas from combustion in an enclosure
        
        Parameters
        ----------
        enclosure: object 
            object that contains information about the enclosure including the volume
        mass : float
           mass of combustible gas in enclosure

        Returns
        -------
        P : float
           pressure upon expansion
        '''
        Vol_total = enclosure.V
        Vol_gas   = mass/self.rho_reac
        
        X_reac_stoich = self._Yreac(self.fstoich)[self.reac]*self._MWmix(self._Yreac(self.fstoich))/self.MW[self.reac]
        sigma = ((self._MWmix(self._Yreac(self.fstoich))/self.Treac) /
                 (self._MWmix(self._Yprod(self.fstoich))/self.T_prod(self.fstoich)))
        cp, cv = self.therm[self.reac].get_property(['CPMASS', 'CVMASS'], T = self.Treac, P = self.Preac)
        try:
            gamma = cp/cv
        except:
            gamma = np.nan
        
        VolStoich = Vol_gas/X_reac_stoich

        deltaP  = self.Preac*((((Vol_total+Vol_gas)/Vol_total)*((Vol_total+VolStoich*(sigma-1))/Vol_total))**gamma-1)
        return deltaP



"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""


import numpy as np
from .distributions import LogNormDistribution



class Leak:
    """
    Random release (leak) chance, defined by an orifice diameter and probability of occurrence.

    Attributes
    ----------
    size : float
        Release size, as percentage of total orifice diameter
    mu : float
        Mean of associated normal distribution
    sigma : float
        Standard deviation of associated normal distribution

    """

    def __init__(self, size, mu, sigma):
        if size < 0 or size > 100:
            raise ValueError(f"Provide a valid leak size between 0 and 100% ({size} provided)")
        if mu in [None, "", np.nan] or sigma in [None, "", np.nan]:
            raise ValueError(f"Provide valid values for size {size}% parameters: mu {mu}, sigma {sigma}")

        self.size = size
        self.mu = mu
        self.sigma = sigma

    def get_leak_freq_mean(self):
        leak_freq_dist = LogNormDistribution(mu=self.mu, sigma=self.sigma)
        return leak_freq_dist.mean

    def __str__(self):
        result = f'Leak {self.size}%, mu: {self.mu}, sigma: {self.sigma}'
        return result





"""
Copyright 2015-2023 National Technology & Engineering Solutions of Sandia, LLC (NTESS).
Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.

You should have received a copy of the GNU General Public License along with HyRAM+.
If not, see https://www.gnu.org/licenses/.
"""

import logging
import os

import matplotlib.pyplot as plt
import numpy as np

from . import Fluid, Orifice, Source, Vent, Enclosure, Flame, Jet
from . import IndoorRelease, FuelProperties, Bauwens_method, TNT_method, BST_method
from ._plots import plot_contour
from ..utilities import misc_utils

"""
The API file provides external access to common analysis functions within the physics module.
API functions prep logging, clean up parameters, and return results or error notifications.

Notes:
    * Queries from the C# GUI should call the c_api and not this file to ensure adequate pre-processing occurs
    * Logging should be set up by the user, not the app. The only exception to this is the C API, which sets up logging
    for the GUI requests.
"""

log = logging.getLogger(__name__)


def create_fluid(species, temp=None, pres=None, density=None, phase=None):
    """
    Create fluid from given parameters. Exactly two of temperature, pressure, density, phase must be provided.

    Parameters
    ----------
    species : str or dict
        Name of fluid or dict of blend mixture of format {fluid-name: mole-frac...}

    pres : float or None
        Fluid pressure (Pa).

    temp : float or None
        Fluid temperature (K).

    density : float or None
        Fluid density (kg/m3)

    phase : str or None
        Fluid phase; values from CoolProp spec. If none, Coolprop determines it.
        GUI passes in {'none', 'gas', 'liquid'} where:
            * 'gas' - saturated vapor.
            * 'liquid' - saturated liquid.
            * 'none' or any other string represents gas phase and will be converted to None.

    Returns
    -------
    Fluid object described by provided parameters

    """
    # phase parameter determines whether temperature is used, but temp may still be passed so must clear it.
    phase = misc_utils.parse_phase_key(phase)
    if phase is not None:
        temp = None

    # User should pass only two specifiers for each fluid, after parsing
    num_params = len([x for x in [temp, pres, density, phase] if x is not None])
    if not num_params == 2:
        message = ('Fluid must be defined by exactly two of the following parameters:'
                   ' temperature, pressure, density, phase')
        raise ValueError(message)

    fluid = Fluid(species=species, T=temp, P=pres, rho=density, phase=phase)
    return fluid


def compute_mass_flow(fluid, orif_diam, amb_pres=101325,
                      is_steady=True, tank_vol=None, dis_coeff=1, output_dir=None, create_plot=True):
    """
    Calculate mass flow rate based on given conditions.

    Parameters
    ----------
    fluid : Fluid
        Release fluid object

    orif_diam : float
        Orifice diameter (m).

    amb_pres : float, optional
        Ambient fluid pressure (Pa).

    is_steady : bool
        Whether system is at steady-state.

    tank_vol : float
        Volume of source (tank) (m^3).

    dis_coeff : float
        Discharge coefficient to account for non-plug flow (always <=1, assumed to be 1 for plug flow).

    output_dir : str
        Path to directory in which to place output file(s).

    create_plot : bool
        Whether mass flow vs. time plot should be created

    Returns
    ----------
    result : dict
        If is_steady, only mass flow rate will be returned as single value.
        If is_steady is false, all parameters will be returned and mass_flow_rate will be list of values.
        rates : list of floats
            Mass flow rate(s) (kg/s)
        time_to_empty : float
            Time (s) it takes to blowdown the tank to empty.
        plot : str
            Path to plot of mass flow rate vs. time. Only created if Steady is false.
        times : array of floats
            Times at which mass flow rates occur during blowdown.

    """
    log.info("Mass Flow analysis requested")

    result = {"time_to_empty": None, "plot": "", "times": None, "rates": None}

    orif = Orifice(orif_diam, Cd=dis_coeff)

    if is_steady:
        result['rates'] = [orif.mdot(orif.flow(fluid, amb_pres))]

    else:
        source = Source(tank_vol, fluid)
        mdots, fluid_list, t, sol = source.empty(orif, amb_pres)

        if create_plot:
            if output_dir is None:
                output_dir = os.getcwd()
            filename = "time-to-empty-{}.png".format(misc_utils.get_now_str())
            filepath = os.path.join(output_dir, filename)
            fig = source.plot_time_to_empty()
            fig.savefig(filepath, bbox_inches='tight')
            plt.close(fig)
            result['plot'] = filepath

        result["time_to_empty"] = t[-1]
        result["times"] = t
        result["rates"] = mdots

    log.info("Mass Flow analysis complete")
    return result


def compute_tank_mass_param(species, phase=None, temp=None, pres=None, vol=None, mass=None):
    """
    Tank mass calculation.
    Three of temp (or phase), pressure, volume, or mass are required.

    species : str
        Fluid species formula or name
        (see CoolProp documentation)

    phase : None or str
        CoolProp specifier for phase
        ('liquid' for saturated liquid,
        'gas' for saturated vapor,
        or None to specify temperature and pressure)

    temp : float or None
        Fluid temperature (K)

    pres : float or None
        Fluid absolute pressure (Pa)

    vol : float or None
        Fluid volume (m^3)

    mass : float or None
        Fluid mass (kg)

    Returns
    -------
    result1 : float
        Fluid temperature, pressure, volume, or mass, depending on provided input parameters.
        If saturated phase, first parameter is pressure, volume, or mass.

    result2 : float or None
        If unsaturated, None returned.
        If saturated phase, this will be temperature.
    """
    log.info("Tank Mass calculation requested")

    phase = misc_utils.parse_phase_key(phase)
    if (temp is not None or phase is not None) and pres is not None and vol is not None and mass is not None:
        msg = 'Too many inputs provided - three of [temperature (or phase), pressure, volume, mass] required'
        raise ValueError(msg)

    if vol is not None and mass is not None:
        density = mass / vol  # kg/m3
    else:
        density = None

    fluid = create_fluid(species, temp, pres, density, phase)

    if density is None:
        if vol is None:
            result1 = mass / fluid.rho
        else:
            result1 = vol * fluid.rho
    elif pres is None:
        result1 = fluid.P
    else:
        result1 = fluid.T

    if phase is not None:  # saturated phase
        result2 = fluid.T
    else:
        result2 = None

    log.info("Tank Mass calculation complete")
    return result1, result2


def compute_thermo_param(species, phase=None, temp=None, pres=None, density=None):
    """
    Calculates temperature, pressure or density of species.
    Requires two of [pressure, density, and (temperature or phase)]
    If unsaturated phase, this returns whichever parameter was not provided.
    If saturated phase, this returns temperature and whichever other parameter was not provided.

    Parameters
    ----------
    species : str
        Fluid species formula or name
        (see CoolProp documentation)

    phase : None or str
        CoolProp specifier for phase
        ('liquid' for saturated liquid,
        'gas' for saturated vapor,
        or None to specify temperature and pressure)

    temp : float or None
        Fluid temperature (K)

    pres : float or None
        Fluid pressure (Pa)

    density : float or None
        Fluid density (kg/m^3)

    Returns
    -------
    float
        Fluid temperature, pressure, or density, depending on provided input parameters.
        If saturated phase, first parameter is either density or pressure.

    float or None
        If unsaturated, None returned.
        If saturated phase, this will be temperature.
    """
    log.info("TPD Parameter calculation requested")

    phase = misc_utils.parse_phase_key(phase)
    if pres is not None and density is not None and (temp is not None or phase is not None):
        msg = 'Too many inputs provided - two of [temperature (or phase), pressure, density] are required'
        raise ValueError(msg)

    fluid = create_fluid(species, temp, pres, density, phase)

    result1 = None
    result2 = None

    if phase is not None:
        # saturated phase
        result1 = fluid.rho if density is None else fluid.P
        result2 = fluid.T

    else:
        if density is None:
            result1 = fluid.rho
        elif pres is None:
            result1 = fluid.P
        else:
            result1 = fluid.T

    log.info("TPD Parameter calculation complete")
    return result1, result2


def compute_equivalent_tnt_mass(vapor_mass, percent_yield, species):
    """
    Calculate equivalent mass of TNT.

    Parameters
    ----------
    vapor_mass : float
        Mass of flammable vapor released (kg)

    percent_yield : float
        Explosive energy yield (0 to 100%)
    
    species : str or dict
        Name of fluid or dict of blend mixture of format {fluid-name: mole-frac...}

    Returns
    ----------
    equiv_TNT_mass : float
        Equivalent mass of TNT (kg)
    """
    if type(species) == dict:
        parsed_dict = misc_utils.parse_blend_dict_into_coolprop_dict(species)
        species = misc_utils.parse_coolprop_dict_into_string(parsed_dict)

    heat_of_combustion = FuelProperties(species).dHc  # J/kg
    log.info(f"TNT mass calculation: vapor mass {vapor_mass:,.4f} kg, yield {percent_yield:,.1f}%, heat of combustion {heat_of_combustion:,.5f} J/kg")
    yield_fraction = percent_yield / 100
    result = TNT_method.calc_TNT_equiv_mass(equivalence_factor=yield_fraction,
                                            flammable_mass=vapor_mass,
                                            heat_of_combustion=heat_of_combustion)
    log.info("TNT mass calculation complete")
    return result


def analyze_jet_plume(amb_fluid, rel_fluid, orif_diam, mass_flow=None,
                      rel_angle=(np.pi/2), dis_coeff=1, nozzle_model='yuce',
                      create_plot=True, contours=None,
                      xmin=None, xmax=None, ymin=None, ymax=None, vmin=0, vmax=0.1,
                      plot_title="Mole Fraction of Leak",
                      filename=None, output_dir=None, verbose=False):
    """
    Simulate jet plume for leak and generate plume positional data, including mass and mole fractions, plume plot.

    Parameters
    ----------
    amb_fluid : Fluid
        Ambient fluid object

    rel_fluid : Fluid
        Release fluid object

    orif_diam : float
        Diameter of orifice (m).

    mass_flow : float or None
        fluid flow rate when unchoked [kg/s]

    rel_angle : float
        Angle of release (radian). 0 is horizontal, pi/2 is vertical.

    dis_coeff : float
        Release discharge coefficient (unitless).

    nozzle_model: {'yuce', 'hars', 'ewan', 'birc', 'bir2', 'molk'}
        Notional nozzle model id. Will be parsed to ensure str matches available options.

    create_plot : bool
        Whether mole fraction contour plot should be created

    contours : list or int or float
        Define contour lines
        Default is None: will use default LFL for selected fuel

    xmin : float, optional
        Plot x-axis minimum.

    xmax : float, optional
        Plot x-axis maximum.

    ymin : float, optional
        Plot y-axis minimum.

    ymax : float, optional
        Plot y-axis maximum.

    vmin : float, optional
        Molar fraction (color bar) scale minimum

    vmax : float, optional
        Molar fraction (color bar) scale maximum

    plot_title : str
        Title displayed in output plot.

    filename : str or None
        Plot filename, excluding path.

    output_dir : str or None
        Directory in which to place plot file.

    verbose : bool, False

    Returns
    -------
    result_dict : dict
        plot : str
            plot file path
        xs : 2D ndarray
            horizontal positions (m) along jet plume
        ys : 2D ndarray
            vertical positions (m) along jet plume
        mole_fracs : 2D ndarray
            mole fractions along jet plume
        mass_fracs : 2D ndarray
            mass fractions along jet plume
        vs : 2D ndarray
            velocities along jet plume
        temps : 2D ndarray
            temperatures (K) along jet plume
        mass_flow_rate : float
            Mass flow rate (kg/s) of steady release.
        streamline_dists: [float]
            Distance(s) of streamline for each contour
        mole_frac_dists: dict
            x and y min-max values for each contour. {contour (float): [(x1, x2), (y1, y2) }
    """
    log.info("Plume plot requested")
    if contours is None:
        fuel_props = FuelProperties(rel_fluid.species)
        contours = [fuel_props.LFL]

    contours = [contours] if type(contours) in [int, float] else contours
    contours = np.sort(contours)

    log.info('Creating components')
    orifice = Orifice(orif_diam, dis_coeff)

    # Jet requires parameters defining notional nozzle model, rather than model itself.
    nozzle_cons_momentum, nozzle_t_param = misc_utils.convert_nozzle_model_to_params(nozzle_model, rel_fluid)

    log.info('Creating jet')
    jet_obj = Jet(rel_fluid, orifice, amb_fluid, theta0=rel_angle, mdot=mass_flow,
                  nn_conserve_momentum=nozzle_cons_momentum,
                  nn_T=nozzle_t_param, verbose=verbose)

    xs, ys, mole_fracs, mass_fracs, vs, temps = jet_obj.get_contour_data()
    mass_flow_rate = jet_obj.mass_flow_rate
    streamline_dists = jet_obj.get_streamline_distances_to_mole_fractions(contours)
    mole_frac_dists = jet_obj.get_xy_distances_to_mole_fractions(contours)

    if create_plot:
        log.info("Creating mole fraction contour plot")
        if output_dir is None:
            output_dir = misc_utils.get_temp_folder()

        if filename is None:
            filename = "plume-mole-plot-{}.png".format(misc_utils.get_now_str())

        if xmin is not None and xmax is not None:
            if xmin >= xmax:
                raise ValueError('x minimum must be less than x maximum')
            xlims = (xmin, xmax)
        else:
            xlims = None

        if ymin is not None and ymax is not None:
            if ymin >= ymax:
                raise ValueError('y minimum must be less than y maximum')
            ylims = (ymin, ymax)
        else:
            ylims = None

        if vmin is None:
            vmin = 0
        if vmax is None:
            vmax = 0.1

        plot_filepath = os.path.join(output_dir, filename)
        plot_fig = plot_contour(data_type="mole", jet_or_flame=jet_obj,
                                xlims=xlims, ylims=ylims, vlims=(vmin, vmax),
                                plot_title=plot_title, contour_levels=contours,)
        plot_fig.savefig(plot_filepath, bbox_inches='tight')
        plt.close(plot_fig)
        log.info("Plume plot complete")
    else:
        plot_filepath = ''

    result_dict = {'xs': xs,
                   'ys': ys,
                   'mole_fracs': mole_fracs,
                   'mass_fracs': mass_fracs,
                   'vs': vs,
                   'temps': temps,
                   'plot': plot_filepath,
                   'mass_flow_rate': mass_flow_rate,
                   'streamline_dists': streamline_dists,
                   'mole_frac_dists': mole_frac_dists}
    return result_dict


def analyze_accumulation(amb_fluid, rel_fluid,
                         tank_volume, orif_diam, rel_height,
                         enclos_height, floor_ceil_area,
                         ceil_vent_xarea, ceil_vent_height,
                         floor_vent_xarea, floor_vent_height,
                         times, orif_dis_coeff=1, ceil_vent_coeff=1, floor_vent_coeff=1,
                         vol_flow_rate=0, dist_rel_to_wall=np.inf,
                         tmax=None, rel_area=None, rel_angle=0, nozzle_key='yuce',
                         x0=0, y0=0, nmax=1000,
                         temp_pres_points=None, pres_ticks=None, is_steady=False,
                         create_plots=True, output_dir=None, verbose=False):
    """
    Simulate an indoor release of designated fluid.

    Parameters
    ----------
    amb_fluid : Fluid
        Ambient fluid object

    rel_fluid : Fluid
        Release fluid object

    tank_volume : float
        Tank volume (m^3).

    orif_diam : float
        Orifice diameter (m).

    rel_height : float
        Height of release (m) above floor at 0.

    enclos_height : float
        Height of enclosure (m).

    floor_ceil_area : float
        Area of floor and ceiling (m^2).

    ceil_vent_xarea : float
        Cross-sectional area of ceiling vent (m^2).

    ceil_vent_height: float
        Height from floor to middle of ceiling vent (m).

    floor_vent_xarea : float
        Cross-sectional area of floor vent (m^2).

    floor_vent_height: float
        Height from floor to middle of floor vent (m).

    times : list or ndarray
        Times at which to return the overpressure (s).

    orif_dis_coeff : float
        Orifice discharge coefficient [unitless].

    ceil_vent_coeff : float
        Discharge coefficient for ceiling vent.

    floor_vent_coeff : float
        Discharge coefficient for floor vent.

    vol_flow_rate : float
        Volumetric flow rate for vents (m^3/s).

    dist_rel_to_wall : float
        Distance between release and wall (m).

    tmax : float, optional
        Maximum simulation time, must be greater than maximum value in t list (s).

    rel_area : float, optional
        Area of release (m^2).

    rel_angle : float
        Angle of release (0 is horizontal) (radians).

    nozzle_key : {'yuce', 'ewan', 'birc', 'bir2', 'molk'}
        Notional nozzle model identifier (i.e. for under-expanded jet zone)

    x0 : float (optional)
        x-coordinate of release (m), default is 0

    y0 : float (optional)
        y-coordinate of release (m), default is 0

    nmax: int, optional
        maximum number of iterations for blowdown integration

    temp_pres_points : 2xn array, optional
        List of [t, P], pairs to plot data points where pairs are in units of [s, kPa].

    pres_ticks : ndarray, optional
        List of pressures at which to draw horizontal lines.

    create_plots : bool
        Whether plots should be generated. If false, plot path params will contain ''.

    output_dir : str, optional
        Filepath of output directory in which to place plots. Will use cwd if not provided.

    verbose : bool
        Verbosity of logging and print statements.

    Returns
    ----------
    result_dict : dict
        status : bool
            Set to 1 if successful
        pressures_per_time : ndarray of floats
            Pressure values (Pa) per time step.
        depths : ndarray of floats
             Depth locations (m) at which concentrations determined.
        concentrations: ndarray
            Fluid concentrations (%) at specified depths.
        overpressure: float
            Max overpressure (Pa) value.
        time_of_overp: float
            Time (s) at which max over-pressure occurred.
        pres_plot_filepath: str
            file path to pressure plot
        mass_plot_filepath: str
            file path to mass plot
        layer_plot_filepath: str
            file path to layer plot
        trajectory_plot_filepath: str
            file path to trajectory plot
        mass_flow_plot_filepath
            file path to mass flow rate plot

    """
    log.info("Accumulation analysis requested")
    if isinstance(times, list):
        times = np.array(times)

    log.info('Creating components')
    rel_source = Source(tank_volume, rel_fluid)
    orifice = Orifice(orif_diam, orif_dis_coeff)

    conserve_momentum, notional_nozzle_t = misc_utils.convert_nozzle_model_to_params(nozzle_key, rel_fluid)
    ceil_vent = Vent(ceil_vent_xarea, ceil_vent_height, ceil_vent_coeff, vol_flow_rate)
    floor_vent = Vent(floor_vent_xarea, floor_vent_height, floor_vent_coeff, vol_flow_rate)
    enclosure = Enclosure(enclos_height, floor_ceil_area, rel_height, ceil_vent, floor_vent,
                          Xwall=dist_rel_to_wall)

    release_obj = IndoorRelease(rel_source, orifice, amb_fluid, enclosure,
                                tmax=tmax, release_area=rel_area, steady=is_steady,
                                nn_conserve_momentum=conserve_momentum, nn_T=notional_nozzle_t,
                                theta0=rel_angle, x0=x0, y0=y0, nmax=nmax, verbose=verbose)

    # interpolate mass flow rates for given times
    rates = np.interp(times, release_obj.ts, release_obj.mdots)

    # Generate plots
    if create_plots:
        if output_dir is None:
            output_dir = os.getcwd()

        now_str = misc_utils.get_now_str()
        pres_plot_fpath = os.path.join(output_dir, 'pressure_plot_{}.png'.format(now_str))
        layer_plot_fpath = os.path.join(output_dir, 'layer_plot_{}.png'.format(now_str))
        traj_plot_fpath = os.path.join(output_dir, 'trajectory_plot_{}.png'.format(now_str))
        mass_plot_fpath = os.path.join(output_dir, 'flam_mass_plot_{}.png'.format(now_str))
        mass_flow_plot_fpath = os.path.join(output_dir, 'time-to-empty_{}.png'.format(now_str))

        pfig = release_obj.plot_overpressure(temp_pres_points, pres_ticks)
        pfig.savefig(pres_plot_fpath, bbox_inches='tight')
        plt.close(pfig)

        lfig = release_obj.plot_layer()
        lfig.savefig(layer_plot_fpath, bbox_inches='tight')
        plt.close(lfig)

        tfig = release_obj.plot_trajectories()
        tfig.savefig(traj_plot_fpath, bbox_inches='tight')
        plt.close(tfig)

        mfig = release_obj.plot_mass()
        mfig.savefig(mass_plot_fpath, bbox_inches='tight')
        plt.close(mfig)

        if is_steady:
            mass_flow_plot_fpath = ''
        else:
            mdotfig = rel_source.plot_time_to_empty()
            mdotfig.savefig(mass_flow_plot_fpath, bbox_inches = 'tight')
            plt.close(mdotfig)

    else:
        pres_plot_fpath = ''
        layer_plot_fpath = ''
        traj_plot_fpath = ''
        mass_plot_fpath = ''
        mass_flow_plot_fpath = ''

    max_pres, max_time = release_obj.max_p_t()

    result_dict = {
        'status': 1,
        'pressures_per_time': release_obj.pressure(times),
        'depths': release_obj.layer_depth(times),
        'concentrations': release_obj.concentration(times),
        'overpressure': max_pres,
        'time_of_overp': max_time,
        'mass_flow_rates': rates,
        'pres_plot_filepath': pres_plot_fpath,
        'mass_plot_filepath': mass_plot_fpath,
        'layer_plot_filepath': layer_plot_fpath,
        'trajectory_plot_filepath': traj_plot_fpath,
        'mass_flow_plot_filepath': mass_flow_plot_fpath
    }

    log.info("Accumulation analysis complete")
    return result_dict


def jet_flame_analysis(amb_fluid, rel_fluid, orif_diam, mass_flow=None, dis_coeff=1,
                       rel_angle=0,
                       nozzle_key='yuce', rel_humid=0.89,
                       # temperature plot
                       create_temp_plot=True, temp_plot_filename=None, temp_plot_title="", temp_contours=None,
                       temp_xlims=None, temp_ylims=None,
                       # heat flux plot
                       analyze_flux=True, flux_plot_filename=None, flux_coordinates=None, flux_contours=None,
                       flux_xlims=None, flux_ylims=None, flux_zlims=None,
                       output_dir=None, verbose=False):
    """
    Assess jet flame behavior and flux data and create corresponding plots.

    Parameters
    ----------
    amb_fluid : Fluid
        Ambient fluid object

    rel_fluid : Fluid
        Release fluid object

    orif_diam : float
        Orifice diameter (m)

    mass_flow : float or None
        fluid flow rate when unchoked [kg/s]

    dis_coeff : float
        Orifice discharge coeffecient [unitless]

    rel_angle : float
        Angle of release (0 is horizontal) (radians)

    nozzle_key : {'yuce', 'ewan', 'birc', 'bir2', 'molk'}
        Notional nozzle model identifier (i.e. for under-expanded jet zone)

    rel_humid : float
        Relative humidity between 0 and 1

    create_temp_plot : bool, True
        Whether temperature plot should be created

    temp_plot_filename : str or None
        Desired filename of output temp plot file

    temp_plot_title : str
        Title to display in plot

    temp_contours : array-like or None
        temperatures at which to draw contour lines

    temp_xlims : tuple/list or None, optional
        Temperature plot x limits: (x_min, x_max). Default is None to use automatic limits.

    temp_ylims : tuple/list or None, optional
        Temperature plot y limits: (y_min, y_max). Default is None to use automatic limits.

    analyze_flux : bool, True
        Whether radiative heat flux analysis should be performed,
        including creating heat flux plots and generating flux data

    flux_plot_filename : str or None
        Filename of heat flux plot file

    flux_coordinates : list of locations
        List of locations at which to determine flux,
        each location is a tuple of 3 coordinates (m):
        [(x1, y1, z1), (x2, y2, z2), ...]

    flux_contours : array-like or None
        Values at which to plot heat flux contours
        (default: 1.577, 4.732, and 25.237 kW/m2)

    flux_xlims : tuple/list or None, optional
        Heat flux plot x limits: (x_min, x_max). Default is None to use automatic limits.

    flux_ylims : tuple/list or None, optional
        Heat flux plot y limits: (y_min, y_max). Default is None to use automatic limits.

    flux_zlims : tuple/list or None, optional
        Heat flux plot z limits: (z_min, z_max). Default is None to use automatic limits.

    output_dir : str
        Filepath of output directory in which to place plots,
        will use cwd if not provided

    verbose : bool
        Verbosity of logging and print statements

    Returns
    -------
    temp_plot_filepath : str or None
        absolute path to temperature plot file

    heatflux_filepath : str or None
        absolute path to temperature plot file

    pos_flux : ndarray or None
        positional flux data (W/m2)

    mass_flow_rate : float
        mass flow rate (kg/s) of the jet flame

    srad : float
        total emitted radiative power of the jet flame (W)

    visible_length : float
        length of visible flame (m)

    radiant_frac : float
        Radiant fraction of flame

    """
    log.info("Jet flame analysis requested")
    now_str = misc_utils.get_now_str()

    if output_dir is None:
        output_dir = os.getcwd()
    if flux_contours is None:
        flux_contours = [1.577, 4.732, 25.237]

    log.info('Creating components')
    orifice = Orifice(orif_diam, Cd=dis_coeff)

    conserve_momentum, notional_nozzle_t = misc_utils.convert_nozzle_model_to_params(nozzle_key, rel_fluid)
    flame_obj = Flame(rel_fluid, orifice, amb_fluid,
                      theta0=rel_angle, y0=0, mdot=mass_flow,
                      nn_conserve_momentum=conserve_momentum, nn_T=notional_nozzle_t,
                      verbose=verbose)

    mass_flow = flame_obj.mass_flow_rate
    srad = flame_obj.get_srad()
    radiant_frac = flame_obj.Xrad
    visible_length = flame_obj.get_visible_length()

    if create_temp_plot:
        log.info("Creating temp plot")
        if temp_plot_filename is None:
            temp_plot_filename = 'flame_temp_plot_{}.png'.format(now_str)
        temp_plot_filepath = os.path.join(output_dir, temp_plot_filename)

        fig = plot_contour(data_type="temperature", jet_or_flame=flame_obj, xlims=temp_xlims, ylims=temp_ylims, plot_title=temp_plot_title,
                           contour_levels=temp_contours)
        fig.savefig(temp_plot_filepath, bbox_inches='tight')
        plt.close(fig)
    else:
        log.info("Skipping temp plot")
        temp_plot_filepath = None

    if analyze_flux:
        log.info("Assessing flux and creating plots")
        if flux_plot_filename is None:
            flux_plot_filename = 'flame_heatflux_{}.png'.format(now_str)

        heatflux_filepath = flame_obj.plot_heat_flux_sliced(filename=flux_plot_filename,
                                                            directory=output_dir,
                                                            RH=rel_humid, contours=flux_contours,
                                                            xlims=flux_xlims, ylims=flux_ylims, zlims=flux_zlims,
                                                            savefig=True)
        pos_flux = flame_obj.generate_positional_flux(flux_coordinates, rel_humid)
    else:
        log.info("Skipping flux analysis")
        heatflux_filepath = None
        pos_flux = None

    log.info("Jet flame analysis complete")
    return temp_plot_filepath, heatflux_filepath, pos_flux, mass_flow, srad, visible_length, radiant_frac


def compute_overpressure(method: str, locations,
                         ambient_fluid, release_fluid,
                         orifice_diameter: float, mass_flow=None,
                         release_angle: float = 0, discharge_coefficient: float = 1,
                         nozzle_model: str = 'yuce',
                         bst_flame_speed: float = 0.35, tnt_factor: float = 0.03,
                         flammability_limits=None,
                         origin_at_orifice=False,
                         create_overpressure_plot: bool = True,
                         create_impulse_plot: bool = True,
                         overpressure_plot_filename=None,
                         impulse_plot_filename=None,
                         output_dir=None, verbose=False,
                         overp_contours=None,
                         overp_xlims=None, overp_ylims=None, overp_zlims=None,
                         impulse_contours=None,
                         impulse_xlims=None, impulse_ylims=None, impulse_zlims=None,
                         ):
    """
    Calculate the overpressure and impulse at a specified locations

    Parameters
    ----------
    method : {'bst', 'tnt', 'bauwens'}
        Unconfined overpressure calculation method
    
    locations : list
        List of locations at which to determine flux,
        each location is a tuple of 3 coordinates (m):
        [(x1, y1, z1), (x2, y2, z2), ...]

    ambient_fluid : Fluid
        Ambient fluid object

    release_fluid : Fluid
        Release fluid object

    orifice_diameter : float
        Diameter of orifice (m)

    mass_flow : float or None
        fluid flow rate when unchoked [kg/s]

    release_angle : float
        Angle of release (radian). 0 is horizontal, pi/2 is vertical.

    discharge_coefficient : float
        Release discharge coefficient (unitless)

    nozzle_model: {'yuce', 'hars', 'ewan', 'birc', 'bir2', 'molk'}
        Notional nozzle model id. Will be parsed to ensure str matches available options.

    bst_flame_speed : float, optional, only needed for BST model
        available mach flame speeds 0.2, 0.35, 0.7, 1.0, 1.4, 2.0, 3.0, 4.0, 5.2
        use 5.2 for detonation 

    tnt_factor : float, optional, only needed for TNT model
        equivalence factor, float
        TNT equivalency, unitless
        # based on HSE guidance in CCPS book https://ebookcentral.proquest.com/lib/sandia/detail.action?docID=624492
        equivalence_factor = 0.03
    
    flammability_limits : tuple of (lower_flammability_limit, upper_flammability_limit), optional
        Default is None: use default LFL, UFL for selected fuel
        
    origin_at_orifice : boolean, optional, default to False
        specify if the origin should be at the orifice or calculated

    create_overpressure_plot : bool, True
        Whether overpressure plot should be created

    create_impulse_plot : bool, True
        Whether impulse plot should be created

    overpressure_plot_filename : str or None
        name of overpressure plot file

    impulse_plot_filename : str or None
        name of impulse plot file

    output_dir : str
        Filepath of output directory in which to place plots. Will use cwd if not provided.

    overp_contours : array-like or None
        Overpressure values at which to plot contours

    overp_xlims : tuple/list, None, optional
        Overpressure plot x limits: (x_min, x_max). Default is None to use automatic limits.

    overp_ylims : tuple/list, None, optional
        Overpressure plot y limits: (y_min, y_max). Default is None to use automatic limits.

    overp_zlims : tuple/list, None, optional
        Overpressure plot z limits: (z_min, z_max). Default is None to use automatic limits.

    impulse_contours : array-like or None
        Impulse values at which to plot contours

    impulse_xlims : tuple/list, None, optional
        Impulse plot x limits: (x_min, x_max). Default is None to use automatic limits.

    impulse_ylims : tuple/list, None, optional
        Impulse plot y limits: (y_min, y_max). Default is None to use automatic limits.

    impulse_zlims : tuple/list, None, optional
        Impulse plot z limits: (z_min, z_max). Default is None to use automatic limits.

    verbose : bool, False

    Returns
    -------
    dict
        overpressures : ndarray of floats
            overpressure in Pa
        impulses : ndarray of floats
            impulse in Pa*s
        overp_plot_filepath : str or None
            path to overpressure figure
        impulse_plot_filepath : str or None
            path to impulse figure
        mass_flow_rate : float
            Mass flow rate (kg/s) of release
        flam_or_det_mass : float
            Mass of release related to overpressure calculation (kg). 
            Outputs flammable mass for BST and TNT methods, detonable mass for Bauwens

    """
    log.info("Unconfined overpressure calculation requested")
    params = locals()
    log.info(misc_utils.params_as_str(params))

    orifice = Orifice(orifice_diameter, discharge_coefficient)
    nozzle_cons_momentum, nozzle_t_param = misc_utils.convert_nozzle_model_to_params(nozzle_model, release_fluid)
    log.info('Creating jet')
    jet_object = Jet(release_fluid, orifice, ambient_fluid, mdot=mass_flow, theta0=release_angle,
                     nn_conserve_momentum=nozzle_cons_momentum, nn_T=nozzle_t_param, verbose=verbose)

    method = method.lower()
    if method == 'bst':
        log.info("Calculating overpressure with BST method")
        over_pressure_model = BST_method(jet_object=jet_object,
                                         mach_flame_speed=bst_flame_speed,
                                         flammability_limits=flammability_limits,
                                         origin_at_orifice=origin_at_orifice)
        flam_or_det_mass = over_pressure_model.flammable_mass
    elif method == 'tnt':
        log.info("Calculating overpressure with TNT method")
        over_pressure_model = TNT_method(jet_object=jet_object,
                                         equivalence_factor=tnt_factor,
                                         flammability_limits=flammability_limits,
                                         origin_at_orifice=origin_at_orifice)
        flam_or_det_mass = over_pressure_model.flammable_mass
    elif method == 'bauwens':
        log.info("Calculating overpressure with Bauwens method")
        over_pressure_model = Bauwens_method(jet_object=jet_object,
                                             origin_at_orifice=origin_at_orifice)
        flam_or_det_mass = over_pressure_model.detonable_mass
    else:
        raise ValueError('Invalid method name in unconfined overpressure analysis')
    
    overpressure = over_pressure_model.calc_overpressure(locations)
    impulse = over_pressure_model.calc_impulse(locations)

    if create_overpressure_plot:
        log.info("Creating overpressure plot")
        if output_dir is None:
            output_dir = os.getcwd()
        if overpressure_plot_filename is None:
            now_str = misc_utils.get_now_str()
            overpressure_plot_filename = 'overpressure_plot_{}.png'.format(now_str)
        overpressure_plot_filepath = over_pressure_model.plot_overpressure_sliced(
                                                            plot_filename=overpressure_plot_filename,
                                                            directory=output_dir,
                                                            contours=overp_contours,
                                                            xlims=overp_xlims,
                                                            ylims=overp_ylims,
                                                            zlims=overp_zlims,
                                                            savefig=True)
    else:
        log.info("Skipping overpressure plot")
        overpressure_plot_filepath = None

    if create_impulse_plot and not np.isnan(impulse).any():
        log.info("Creating impulse plot")
        if output_dir is None:
            output_dir = os.getcwd()
        if impulse_plot_filename is None:
            now_str = misc_utils.get_now_str()
            impulse_plot_filename = 'impulse_plot_{}.png'.format(now_str)
        impulse_plot_filepath = over_pressure_model.plot_impulse_sliced(
                                                            plot_filename=impulse_plot_filename,
                                                            directory=output_dir,
                                                            contours=impulse_contours,
                                                            xlims=impulse_xlims,
                                                            ylims=impulse_ylims,
                                                            zlims=impulse_zlims,
                                                            savefig=True)
    else:
        log.info("Skipping impulse plot")
        impulse_plot_filepath = None

    log.info("Unconfined overpressure calculation complete")
    results = {
        'overpressures': overpressure,
        'impulses': impulse,
        'overp_plot_filepath': overpressure_plot_filepath,
        'impulse_plot_filepath': impulse_plot_filepath,
        'mass_flow_rate': jet_object.mass_flow_rate,
        'flam_or_det_mass': flam_or_det_mass
    }
    return results
